# -*- coding: utf-8 -*-
"""
Created on Tue Oct 13 11:19:36 2015

@author: erlean
"""
__tissues = {1: 'hardbone',
               2: 'skin',
               3: 'muscle',
               4: 'soft',
               5: 'redbonemarrow',
               6: 'yellowbonemarrow',
               7: 'adipose',
               8: 'lung',
               9: 'air',
               10: 'water',
               }

__jo_organs = """1;Adrenal, left;4
                2;Adrenal, right;4
                3;Anterior nasal passage (ET1);9
                4;Posterior nasal passage down to larynx (ET2);4
                5;Oral mucosa, tongue;3
                6;Oral mucosa, lips and cheeks;7
                7;Trachea;4
                8;Bronchi;8
                9;Blood vessels, head;5
                10;Blood vessels, trunk;5
                11;Blood vessels, arms;5
                12;Blood vessels, legs;5
                13;Humeri, proximal end, cortical bone;1
                14;Humeri, upper half, spongiosa;1
                15;Humeri, upper half, medullary cavity;1
                16;Humeri, lower half, cortical;1
                17;Humeri, lower half, spongiosa;1
                18;Humeri, lower half, medullary cavity;1
                19;Ulnae and radii, cortical;1
                20;Ulnae and radii, spongiosa;1
                21;Ulnae and radii, medullary cavity;1
                22;Wrists and hand bones, cortical;1
                23;Wrists and hand bones, spongiosa;1
                24;Clavicles, cortical;1
                25;Clavicles, spongiosa;1
                26;Cranium, cortical;1
                27;Cranium, spongiosa;1
                28;Femora, upper half, cortical;1
                29;Femora, upper half, spongiosa;1
                30;Femora, upper half, medullary cavity;1
                31;Femora, lower half, cortical;1
                32;Femora, lower half, spongiosa;1
                33;Femora, lower half, medullary cavity;1
                34;Tibiae, fibulae and patellae:;1
                35;Tibiae, fibulae and patellae, spongiosa;1
                36;Tibiae, fibulae and patellae, medullary cavity;1
                37;Ankles and foot bones:;1
                38;Ankles and foot bones, spongiosa;1
                39;Mandible, cortical;1
                40;Mandible, spongiosa;1
                41;Pelvis, cortical;1
                42;Pelvis, spongiosa;1
                43;Ribs, cortical;1
                44;Ribs, spongiosa;1
                45;Scapulae, cortical;1
                46;Scapulae, spongiosa;1
                47;Cervical spine, cortical;1
                48;Cervical spine, spongiosa;1
                49;Thoracic spine, cortical;1
                50;Thoracic spine, spongiosa;1
                51;Lumbar spine, cortical;1
                52;Lumbar spine, spongiosa;1
                53;Sacrum, cortical;1
                54;Sacrum, spongiosa;1
                55;Sternum, cortical;1
                56;Sternum, spongiosa;1
                57;Cartilage, head;1
                58;Cartilage, trunk;1
                59;Cartilage, arms;1
                60;Cartilage, legs;1
                61;Brain;7
                62;Breast, left, adipose tissue;7
                63;Breast, left, glandular tissue;4
                64;Breast, right, adipose tissue;7
                65;Breast, right, glandular tissue;4
                66;Eye lens, left;4
                67;Eye bulb, left;4
                68;Eye lens, right;4
                69;Eye bulb, right;4
                70;Gall bladder wall;4
                71;Gall bladder contents;10
                72;Stomach wall;4
                73;Stomach contents;4
                74;Small intestine wall;4
                75;Small intestine contents;4
                76;Ascending colon wall;4
                77;Ascending colon contents;4
                78;Transverse colon wall, right;4
                79;Transverse colon contents, right;4
                80;Transverse colon wall, left;4
                81;Transverse colon contents, left;4
                82;Descending colon wall;4
                83;Descending colon contents;4
                84;Sigmoid colon wall;4
                85;Sigmoid colon contents;4
                86;Rectum wall;4
                87;Heart wall;3
                88;Heart contents (blood);5
                89;Kidney, left, cortex;4
                90;Kidney, left, medulla;4
                91;Kidney, left, pelvis;4
                92;Kidney, right, cortex;4
                93;Kidney, right, medulla;4
                94;Kidney, right, pelvis;4
                95;Liver;4
                96;Lung, left, blood;5
                97;Lung, left, tissue;8
                98;Lung, right, blood;5
                99;Lung, right, tissue;8
                100;Lymphatic nodes, extrathoracic airways;4
                101;Lymphatic nodes, thoracic airways;4
                102;Lymphatic nodes, head;4
                103;Lymphatic nodes, trunk;4
                104;Lymphatic nodes, arms;4
                105;Lymphatic nodes, legs;4
                106;Muscle, head;3
                107;Muscle, trunk;3
                108;Muscle, arms;3
                109;Muscle, legs;3
                110;Oesophagus (wall);3
                111;Ovary, left;4
                112;Ovary, right;4
                113;Pancreas;4
                114;Pituitary gland;4
                115;Prostate;4
                116;Residual tissue, head;7
                117;Residual tissue, trunk;7
                118;Residual tissue, arms;7
                119;Residual tissue, legs;7
                120;Salivary glands, left;4
                121;Salivary glands, right;4
                122;Skin, head;2
                123;Skin, trunk;2
                124;Skin, arms;2
                125;Skin, legs;2
                126;Spinal cord;6
                127;Spleen;4
                128;Teeth;1
                129;Testis, left;4
                130;Testis, right;4
                131;Thymus;4
                132;Thyroid;4
                133;Tongue (inner part);3
                134;Tonsils;4
                135;Ureter, left;4
                136;Ureter, right;4
                137;Urinary bladder wall;4
                138;Urinary bladder contents;4
                139;Uterus / Cervix;4
                140;Air inside body;9
                141;Skin at top and bottom;2"""


def jo_organs():
    lg = __jo_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]


__baby_organs = """0;Air;9
                    1;Adrenals;4
                    2;Bladder wall;4
                    3;Brain;7
                    4;Eyes;4
                    5;Eye lenses;4
                    6;Heart;3
                    8;Small intestine wall;4
                    10;Left kidney;4
                    11;Right kidney;4
                    12;Liver;4
                    14;Left lung;8
                    15;Right lung;8
                    16;Ovaries;4
                    17;Pancreas;4
                    19;Head skin;2
                    20;Trunk skin;2
                    21;Leg skin;2
                    22;Spinal cord;4
                    23;Spleen;4
                    24;Stomach wall;4
                    25;Testes;4
                    26;Thymus;4
                    27;Thyroid;4
                    29;Head tissue;4
                    30;Trunk tissue;4
                    31;Leg tissue;4
                    32;Uterus;4
                    34;Upper left arm bone;1
                    35;Lower left arm bones;1
                    36;Left hand bones;1
                    37;Upper right arm bone;1
                    38;Lower right arm bones;1
                    39;Right hand bones;1
                    40;Clavicles;1
                    41;Facial skeleton;1
                    43;Upper left leg bone;1
                    44;Lower left leg bones;1
                    45;Left foot bones;1
                    46;Upper right leg bone;1
                    47;Lower right leg bones;1
                    48;Right foot bones;1
                    49;Pelvis;1
                    50;Ribs;1
                    51;Scapulae;1
                    52;Skull;1
                    54;Cervical spine;1
                    55;Thoracic spine;1
                    56;Lumbar spine;1
                    57;Sternum;1
                    93;Stomach contents;4
                    94;Small intestine contents;4
                    95;UknownERROR;4
                    96;Bladder contents;10
                    97;UknownERROR;4
                    103;Upper large intestine (ascending+transverse colon) wall;4
                    104;Lower large intestine (descending+sigmoid colon) wall;4
                    105;Upper large intestine contents;4
                    106;Lower large intestine contents;4
                    107;Arm skin;2
                    108;Arm tissue;4
                    109;Breast;7
                    110;Oesophagus;4
                    111;Gall bladder wall;4
                    112;Gall bladder contents;4
                    113;Muscle tissue, head;3
                    114;Muscle tissue, trunk;3
                    115;Muscle tissue, arms;3
                    116;Muscle tissue, legs;3"""
def baby_organs():
    lg = __baby_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]

__child_organs="""0;Air;9
                    1;Adrenals;4
                    2;Bladder wall;4
                    3;Brain;7
                    4;Eyes;4
                    5;Eye lenses;4
                    6;Heart;3
                    8;Colon ascending + transverse (wall);4
                    9;Colon descending (wall);4
                    10;Small intestine (wall + contents);4
                    12;Left kidney;4
                    13;Right kidney;4
                    14;Liver;4
                    16;Left lung;8
                    17;Right lung;8
                    18;Ovaries;4
                    19;Pancreas;4
                    21;Head skin;2
                    22;Trunk skin;2
                    23;Arm skin;2
                    24;Leg skin;2
                    25;Spinal cord;4
                    26;Spleen;4
                    27;Stomach wall;4
                    28;Thymus;4
                    29;Thyroid;4
                    31;Head tissue;4
                    32;Trunk tissue;4
                    33;Left arm tissue;3
                    34;Right arm tissue;3
                    35;Left leg tissue;3
                    36;Right leg tissue;3
                    37;Trachea;4
                    38;Uterus;4
                    40;Upper left arm bone;1
                    41;Lower left arm bones;1
                    42;Left hand bones;1
                    43;Upper right arm bone;1
                    44;Lower right arm bones;1
                    45;Right hand bones;1
                    46;Clavicles;1
                    47;Facial skeleton;1
                    49;Upper left leg bone;1
                    50;Lower left leg bones;1
                    51;Left foot bones;1
                    52;Upper right leg bone;1
                    53;Lower right leg bones;1
                    54;Right foot bones;1
                    55;Pelvis;1
                    56;Ribs;1
                    61;Scapulae;1
                    62;Skull;1
                    64;Cervical spine;1
                    65;Thoracic spine;1
                    66;Lumbar spine;1
                    67;Sacrum;1
                    68;Sternum;1
                    110;Mucous membrane (head);4
                    111;Bladder contents;10
                    112;Colon contents;4
                    113;Stomach contents;4
                    114;Testes;4"""
def child_organs():
    lg = __child_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]

__katja_organs="""0;Air;9
                1;Adrenal, left;4
                2;Adrenal, right;4
                3;Anterior nasal passage (ET1);2
                4;Posterior nasal passage down to larynx (ET2);2
                5;Oral mucosa, tongue;4
                6;Oral mucosa, lips and cheeks;4
                7;Trachea;4
                8;Bronchi;4
                9;Blood vessels, head;4
                10;Blood vessels, trunk;4
                11;Blood vessels, arms;4
                12;Blood vessels, legs;4
                13;Humeri, proximal end, cortical bone;1
                14;Humeri, upper half, spongiosa;1
                15;Humeri, upper half, medullary cavity;1
                16;Humeri, lower half, cortical;1
                17;Humeri, lower half, spongiosa;1
                18;Humeri, lower half, medullary cavity;1
                19;Ulnae and radii, cortical;1
                20;Ulnae and radii, spongiosa;1
                21;Ulnae and radii, medullary cavity;1
                22;Wrists and hand bones, cortical;1
                23;Wrists and hand bones, spongiosa;1
                24;Clavicles, cortical;1
                25;Clavicles, spongiosa;1
                26;Cranium, cortical;1
                27;Cranium, spongiosa;1
                28;Femora, upper half, cortical;1
                29;Femora, upper half, spongiosa;1
                30;Femora, upper half, medullary cavity;1
                31;Femora, lower half, cortical;1
                32;Femora, lower half, spongiosa;1
                33;Femora, lower half, medullary cavity;1
                34;Tibiae, fibulae and patellae, cortical;1
                35;Tibiae, fibulae and patellae, spongiosa;1
                36;Tibiae, fibulae and patellae, medullary cavity;1
                37;Ankles and foot bones, cortical;1
                38;Ankles and foot bones, spongiosa;1
                39;Mandible, cortical;1
                40;Mandible, spongiosa;1
                41;Pelvis, cortical;1
                42;Pelvis, spongiosa;1
                43;Ribs, cortical;1
                44;Ribs, spongiosa;1
                45;Scapulae, cortical;1
                46;Scapulae, spongiosa;1
                47;Cervical spine, cortical;1
                48;Cervical spine, spongiosa;1
                49;Thoracic spine, cortical;1
                50;Thoracic spine, spongiosa;1
                51;Lumbar spine, cortical;1
                52;Lumbar spine, spongiosa;1
                53;Sacrum, cortical;1
                54;Sacrum, spongiosa;1
                55;Sternum, cortical;1
                56;Sternum, spongiosa;1
                57;Cartilage, head;1
                58;Cartilage, trunk;1
                59;Cartilage, arms;1
                60;Cartilage, legs;1
                61;Brain;4
                62;Breast, left, adipose tissue ;7
                63;Breast, left, glandular tissue;4
                64;Breast, right, adipose tissue;7
                65;Breast, right, glandular tissue;4
                66;Eye lens, left;4
                67;Eye bulb, left;4
                68;Eye lens, right;4
                69;Eye bulb, right;4
                70;Gall bladder wall;4
                71;Gall bladder contents;4
                72;Stomach wall;4
                73;Stomach contents;4
                74;Small intestine wall;4
                75;Small intestine contents;4
                76;Ascending colon wall;4
                77;Ascending colon contents;4
                78;Transverse colon wall, right;4
                79;Transverse colon contents, right;4
                80;Transverse colon wall, left;4
                81;Transverse colon contents, left;4
                82;Descending colon wall;4
                83;Descending colon contents;4
                84;Sigmoid colon wall;4
                85;Sigmoid colon contents;4
                86;Rectum wall;4
                87;Heart wall;3
                88;Heart contents (blood);4
                89;Kidney, left, cortex;4
                90;Kidney, left, medulla;4
                91;Kidney, left, pelvis;4
                92;Kidney, right, cortex;4
                93;Kidney, right, medulla;4
                94;Kidney, right, pelvis;4
                95;Liver;4
                96;Lung, left, blood;4
                97;Lung, left, tissue;8
                98;Lung, right, blood;4
                99;Lung, right, tissue;8
                100;Lymphatic nodes, extrathoracic airways;4
                101;Lymphatic nodes, thoracic airways;4
                102;Lymphatic nodes, head;4
                103;Lymphatic nodes, trunk;4
                104;Lymphatic nodes, arms;4
                105;Lymphatic nodes, legs;4
                106;Muscle, head;3
                107;Muscle, trunk;3
                108;Muscle, arms;3
                109;Muscle, legs;3
                110;Oesophagus (wall);4
                111;Ovary, left;4
                112;Ovary, right;4
                113;Pancreas;4
                114;Pituitary gland;4
                115;Prostate;4
                116;Residual tissue, head;7
                117;Residual tissue, trunk;7
                118;Residual tissue, arms;7
                119;Residual tissue, legs;7
                120;Salivary glands, left;4
                121;Salivary glands, right;4
                122;Skin, head;2
                123;Skin, trunk;2
                124;Skin, arms;2
                125;Skin, legs;2
                126;Spinal cord;2
                127;Spleen;4
                128;Teeth;1
                129;Testis, left;4
                130;Testis, right;4
                131;Thymus;4
                132;Thyroid;4
                133;Tongue (inner part);4
                134;Tonsils;4
                135;Ureter, left;4
                136;Ureter, right;4
                137;Urinary bladder wall;4
                138;Urinary bladder contents;4
                139;Uterus / Cervix;4
                140;Air inside body;9
                141;Skin at top and bottom;2
                210;Placenta;4
                211;Umbilical cord;4
                212;Amniotic fluid;4
                213;Fetus Head tissue;4
                214;Fetus cerebrospinal fluid;10
                215;Fetus Spine;1
                220;Fetus Skin;2
                221;Fetus Trunk;4
                222;Fetus Arms;4
                223;Fetus Legs;4
                224;Fetus Brain;4
                225;Fetus Eyes;4
                226;Fetus Eye lenses;4
                227;Fetus Spinal cord;4
                228;Fetus Lungs;4
                229;Fetus Heart;3
                230;Fetus Kidneys;4
                231;Fetus Liver;4
                232;Fetus Stomach;4
                233;Fetus Gall bladder;4
                240;Fetus Cranium;1"""

def katja_organs():
    lg = __katja_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]


__adam_organs="""0;Air;9
                1;LEFT ADRENAL;4
                2;RIGHT ADRENAL;4
                3;ADRENALS;4
                4;BLADDER;4
                5;CONTENTS;4
                6;BRAIN;7
                7;G.I. STOMACH;4
                8;CONTENTS;4
                9;G.I. U.L.I.;4
                10;CONTENTS;4
                11;G.I. L.L.I.;4
                12;CONTENTS;4
                13;SM.INT.+CONTS;4
                14;GENITALIA;1
                15;HEART;3
                16;LEFT KIDNEY;4
                17;RIGHT KIDNEY;4
                18;KIDNEYS;4
                19;LIVER;4
                20;LEFT LUNG;8
                21;RIGHT LUNG;8
                22;LUNGS;8
                23;RM L.ARM;5
                24;RM R.ARM;5
                25;RM CLAVICLES;5
                26;RM UP.L.LEG;5
                27;RM UP.R.LEG;5
                28;RM PELVIS;5
                29;RM RIBS;5
                30;RM SCAPULAE;5
                31;RM CRANIUM;5
                32;RM FAC.SKEL.;5
                33;RM LOW.SPINE;5
                34;RM MID.SPINE;5
                35;RM UPP.SPINE;5
                36;RM MID.REGION;5
                37;RM LOW.REGION;5
                38;RM HEAD;5
                39;RED MARROW;5
                40;CR UP.L.ARM;1
                42;CR UP.R.ARM;1
                44;CR CLAVICLES;1
                45;CR UP.L.LEG;1
                47;CR UP.R.LEG;1
                49;CR PELVIS;1
                50;CR RIBS;1
                51;CR SCAPULAE;1
                52;CR CRANIUM;1
                53;CR FAC.SKEL.;1
                54;CR LOW.SPINE;1
                55;CR MID.SPINE;1
                56;CR UPP.SPINE;1
                57;CR MID.REGION;1
                58;CR LOW.REGION;1
                59;CR HEAD;1
                60;ORNL   MARROW;6
                64;GABLAD SURF;4
                65;GABLAD CONT;4
                66;GABLAD TOTAL;4
                67;PANCREAS;4
                68;LEFT ARM BONE;1
                69;RT. ARM BONE;1
                70;CLAVICLES;1
                71;LEFT LEG BONE;1
                72;RT. LEG BONE;1
                73;PELVIS;1
                74;RIBS;1
                75;SCAPULAE;1
                76;CRANIUM;1
                77;SPINE;1
                78;SKELETON;1
                79;TRUNK SKIN;2
                80;LEG SKIN;2
                81;HEAD SKIN;2
                82;TOTAL SKIN;2
                83;SPLEEN;4
                84;LEFT TESTIS;4
                85;RIGHT TESTIS;4
                86;TESTES;4
                87;THYMUS;4
                88;THYROID;4
                89;TRUNK TISSUE;4
                90;LEG TISSUE;4
                91;HEAD TISSUE;4
                92;TOTAL TISSUE;4
                93;FAC.SKELETON;1
                94;TRUNK;4
                95;LEGS;4
                96;HEAD;4
                97;TOTAL BODY;4
                98;SKIN BACKSIDE;2
                99;SKIN FRONTAL;2
                100;PERS. DOSE;9
                101;SURFACE ENTR;9
                102;SURFACE EXIT;9
                103;MARR RIBS PO;1
                104;MARR RIBS AN;1
                105;SURFACE ROT.;9
                106;LENSE OF EYE;4
                107;COLON (9+11);4
                108;OESOPHAGUS;4"""


def adam_organs():
    lg = __adam_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]

__eva_organs = """0;Air;9;
                    1;LEFT ADRENAL;4;
                    2;RIGHT ADRENAL;4;
                    3;ADRENALS;4;
                    4;BLADDER;4;
                    5;BLADDER CONTENTS;4;
                    6;BRAIN;7;
                    7;G.I. STOMACH;4;
                    8;CONTENTS;4;
                    9;G.I. U.L.I.;4;
                    10;CONTENTS;4;
                    11;G.I. L.L.I.;4;
                    12;CONTENTS;4;
                    13;SM.INT.+CONTS;4;
                    14;FAC.SKELETON;1;
                    15;HEART;3;
                    16;LEFT KIDNEY;4;
                    17;RIGHT KIDNEY;4;
                    18;KIDNEYS;4;
                    19;LIVER;4;
                    20;LEFT LUNG;8;
                    21;RIGHT LUNG;8;
                    22;LUNGS;8;
                    23;RM L.ARM;5;
                    24;RM R.ARM;5;
                    25;RM CLAVICLES;5;
                    26;RM UP.L.LEG;5;
                    27;RM UP.R.LEG;5;
                    28;RM PELVIS;5;
                    29;RM RIBS;5;
                    30;RM SCAPULAE;5;
                    31;RM CRANIUM;5;
                    32;RM FAC.SKEL.;5;
                    33;RM LOW.SPINE;5;
                    34;RM MID.SPINE;5;
                    35;RM UPP.SPINE;5;
                    39;RED MARROW;5;
                    64;LEFT OVARY;4;
                    65;RIGHT OVARY;4;
                    66;OVARIES;4;
                    67;PANCREAS;4;
                    68;LEFT ARM BONE;1;
                    69;RT. ARM BONE;1;
                    70;CLAVICLES;1;
                    71;LEFT LEG BONE;1;
                    72;RT. LEG BONE;1;
                    73;PELVIS;1;
                    74;RIBS;1;
                    75;SCAPULAE;1;
                    76;CRANIUM;1;
                    77;SPINE;1;
                    78;SKELETON;1;
                    79;TRUNK SKIN;2;
                    80;LEG SKIN;2;
                    81;HEAD SKIN;2;
                    82;TOTAL SKIN;2;
                    83;SPLEEN;4;
                    84;GABLAD SURF;4;
                    85;GABLAD CONT;4;
                    86;GABLAD TOTAL;4;
                    87;THYMUS;4;
                    88;THYROID;4;
                    89;TRUNK TISSUE;4;
                    90;LEG TISSUE;4;
                    91;HEAD TISSUE;4;
                    92;TOTAL TISSUE;4;
                    93;UTERUS;4;
                    94;TRUNK;4;
                    95;LEGS;4;
                    96;HEAD;4;
                    97;TOTAL BODY;4;
                    98;SKIN BACKSIDE;2;
                    99;SKIN FRONTAL;2;
                    100;PERS. DOSE;9;
                    101;SURFACE ENTR;9;
                    102;SURFACE EXIT;9;
                    105;SURFACE ROT.;9;
                    106;FEM.BREAST;7;
                    107;RIGHT BREAST;7;
                    108;LEFT BREAST;7;
                    109;SKIN R.BRE.;2;
                    110;SKIN L.BREA.;2;
                    111;LENSE OF EYE;4;
                    112;COLON (9+11);4;
                    113;OESOPHAGUS;4;"""
def eva_organs():
    lg = __eva_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]

__donna_helga__irene_frank_organs = """0;Air;9
                    1;Adipose tissue arms;7
                    2;Adipose tissue rest of body;7
                    3;Adrenals;4
                    4;Bladder wall;4
                    5;Bladder contents;4
                    6;Blood vessels in arms;4
                    7;Blood vessels in rest of body;4
                    8;Arm bones;1
                    9;Leg bones;1
                    10;Bones in rest of body;1
                    11;Brain;7
                    12;Breast, adipose tissue ;7
                    13;Breast, glandular tissue;4
                    14;Bronchi;4
                    15;Cartilage in arms;4
                    16;Cartilage in rest of body;4
                    17;Caecum;4
                    18;Ascending colon;4
                    19;Transverse colon;4
                    20;Descending colon;4
                    21;Sigmoid colon;4
                    22;Rectum;4
                    23;Caecum contents;4
                    24;Ascending colon contents;4
                    25;Transverse colon contents;4
                    26;Descending colon contents;4
                    27;Sigmoid colon contents;4
                    28;Rectum contents;4
                    29;Eye lenses;4
                    30;Eyes;4
                    31;Anterior nasal passage (ET1);2
                    32;Posterior nasal passage down to larynx (ET2);2
                    33;Gall bladder wall;4
                    34;Gall bladder contents;4
                    35;Heart wall;3
                    36;Heart contents (blood);4
                    37;Kidneys;4
                    38;Liver;4
                    39;Lungs;8
                    40;Muscle in arms;3
                    41;Muscle in rest of body;3
                    42;Oesophagus;4
                    43;Ovaries;4
                    44;Pancreas;4
                    45;Prostate;4
                    46;Salivary glands;4
                    47;Skin at arms;2
                    48;Skin at rest of body;2
                    49;Duodenum;4
                    50;Ileum;4
                    51;Duodenum contents;4
                    52;Ileum contents;4
                    53;Spinal cord;1
                    54;Spleen;4
                    55;Stomach wall;4
                    56;Stomach contents;4
                    57;Teeth;1
                    58;Testes;4
                    59;Thymus;4
                    60;Thyroid;4
                    61;Tongue;4
                    62;Trachea;4
                    63;Ureters;4
                    64;Uterus;4
                    65;Patient table;4"""
def donna_organs():
    lg = __donna_helga__irene_frank_organs.replace('\n', ';').replace('\r', '').split(';')
    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]

def helga_organs():
    for val in donna_organs():
        yield val

def irene_organs():
    for val in donna_organs():
        yield val
def frank_organs():
    for val in donna_organs():
        yield val


__vishum_organs = """0;Air;9;
                     1;Head adipose;7;
                     2;Trunk adipose;7;
                     3;Left arm adipose;7;
                     4;Right arm adipose;7;
                     5;Left leg adipose;7;
                     6;Right leg adipose;7;
                     7;Adrenals;4;
                     8;Bladder wall;4;
                     9;Bladder contents;4;
                     10;Blood vessels;4;
                     11;Grey matter;7;
                     12;White matter;7;
                     13;Ventricles;10;
                     14;Adipose tissue, left breast;7;
                     15;Adipose tissue, right breast;7;
                     16;Glandular tissue, left breast;4;
                     17;Glandular tissue, right breast;4;
                     18;Bronchi;4;
                     19;Ascending colon;4;
                     20;Transverse colon;4;
                     21;Descending colon;4;
                     22;Sigmoid colon;4;
                     23;Rectum;4;
                     24;Colon contents;4;
                     25;Eye lenses;4;
                     26;Eyes;4;
                     27;Anterior nasal passage;2;
                     28;Posterior nasal passage;2;
                     29;Pharynx (nasal part);2;
                     30;Pharynx (oral part);2;
                     31;Larynx;2;
                     32;Gall bladder wall;4;
                     33;Gall bladder contents;4;
                     34;Heart;4;
                     35;Left kidney;4;
                     36;Right kidney;4;
                     37;Liver;4;
                     38;Left lung;8;
                     39;Right lung;8;
                     40;Head muscle;3;
                     41;Trunk muscle;3;
                     42;Left arm muscle;3;
                     43;Right arm muscle;3;
                     44;Left leg muscle;3;
                     45;Right leg muscle;3;
                     46;Oesophagus;4;
                     47;Ovaries;4;
                     48;Pancreas;4;
                     49;Prostate;4;
                     50;Salivary glands;4;
                     51;Upper left arm bone;1;
                     52;Lower left arm bones;1;
                     53;Left hand bones;1;
                     54;Upper right arm bone;1;
                     55;Lower right arm bones;1;
                     56;Right hand bones;1;
                     57;Clavicles;1;
                     58;Upper left leg bone;1;
                     59;Lower left leg bones;1;
                     60;Left foot bones;1;
                     61;Upper right leg bone;1;
                     62;Lower right leg bones;1;
                     63;Left foot bones;1;
                     64;Mandible;1;
                     65;Pelvis;1;
                     66;Left rib 1;1;
                     67;Left rib 2;1;
                     68;Left rib 3;1;
                     69;Left rib 4;1;
                     70;Left rib 5;1;
                     71;Left rib 6 ;1;
                     72;Left rib 7 ;1;
                     73;Left rib 8 ;1;
                     74;Left rib 9 ;1;
                     75;Left rib 10 ;1;
                     76;Left rib 11 ;1;
                     77;Left rib 12 ;1;
                     78;Right rib 1 ;1;
                     79;Right rib 2 ;1;
                     80;Right rib 3 ;1;
                     81;Right rib 4 ;1;
                     82;Right rib 5 ;1;
                     83;Right rib 6 ;1;
                     84;Right rib 7 ;1;
                     85;Right rib 8 ;1;
                     86;Right rib 9 ;1;
                     87;Right rib 10 ;1;
                     88;Right rib 11 ;1;
                     89;Right rib 12 ;1;
                     90;Scapulae;1;
                     91;Skull;1;
                     92;Cervical vertebra 1;1;
                     93;Cervical vertebra 2 ;1;
                     94;Cervical vertebra 3 ;1;
                     95;Cervical vertebra 4 ;1;
                     96;Cervical vertebra 5 ;1;
                     97;Cervical vertebra 6 ;1;
                     98;Cervical vertebra 7 ;1;
                     99;Thoracic vertebra 1 ;1;
                    100;Thoracic vertebra 2 ;1;
                    101;Thoracic vertebra 3 ;1;
                    102;Thoracic vertebra 4 ;1;
                    103;Thoracic vertebra 5 ;1;
                    104;Thoracic vertebra 6 ;1;
                    105;Thoracic vertebra 7 ;1;
                    106;Thoracic vertebra 8 ;1;
                    107;Thoracic vertebra 9 ;1;
                    108;Thoracic vertebra 10 ;1;
                    109;Thoracic vertebra 11 ;1;
                    110;Thoracic vertebra 12 ;1;
                    111;Lumbar vertebra 1;1;
                    112;Lumbar vertebra 2 ;1;
                    113;Lumbar vertebra 3 ;1;
                    114;Lumbar vertebra 4 ;1;
                    115;Lumbar vertebra 5 ;1;
                    116;Sacrum;1;
                    117;Sternum;1;
                    118;Head skin;2;
                    119;Trunk skin;2;
                    120;Left arm skin;2;
                    121;Right arm skin;2;
                    122;Left leg skin;2;
                    123;Right leg skin;2;
                    124;Small intestine wall;4;
                    125;Small intestine contents;4;
                    126;Spinal cord;4;
                    127;Spleen;4;
                    128;Stomach wall;4;
                    129;Stomach contents;4;
                    130;Teeth;1;
                    131;Testes;4;
                    132;Thymus;4;
                    133;Thyroid;4;
                    134;Tongue;4;
                    135;Trachea;4;
                    136;Uterus;4;
                    137;Patient table;4;"""


def vishum_organs():
    lg = __vishum_organs.replace('\n', ';').replace('\r', '').split(';')

    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]


__golem_organs = """0;Air;9
                    2;Adipose tissue head;7
                    3;Adipose tissue trunk;7
                    4;Adipose tissue left arm;7
                    5;Adipose tissue right arm;7
                    6;Adipose tissue left leg;7
                    7;Adipose tissue right leg;7
                    8;Adrenals;4
                    9;Bladder wall;4
                    27;Brain;4
                    28;Anterior nasal passage;2
                    29;Posterior nasal passage;2
                    30;Pharynx (nasal part);2
                    31;Pharynx (oral part);2
                    32;Larynx;2
                    33;Bronchial tree;4
                    34;Eyes;4
                    35;Eye lenses;4
                    36;Heart;4
                    38;Left kidney;4
                    39;Right kidney;4
                    41;Liver;4
                    43;Left lung;8
                    44;Right lung;8
                    46;Muscle tissue head;3
                    47;Muscle tissue trunk;3
                    48;Muscle tissue left arm;3
                    49;Muscle tissue right arm;3
                    50;Muscle tissue left leg;3
                    51;Muscle tissue right leg;3
                    52;Oesophagus;4
                    54;Pancreas;4
                    55;Penis;4
                    135;Upper left arm bone;1
                    136;Lower left arm bones;1
                    137;Left hand bones;1
                    138;Upper right arm bone;1
                    139;Lower right arm bones;1
                    140;Right hand bones;1
                    141;Clavicles;1
                    143;Upper left leg bone;1
                    144;Lower left leg bones;1
                    145;Left foot bones;1
                    146;Upper right leg bone;1
                    147;Lower right leg bones;1
                    148;Right foot bones;1
                    149;Mandible;1
                    150;Pelvis;1
                    152;Rib 1 left side;1
                    153;Rib 2 left side;1
                    154;Rib 3 left side;1
                    155;Rib 4 left side;1
                    156;Rib 5 left side;1
                    157;Rib 6 left side;1
                    158;Rib 7 left side;1
                    159;Rib 8 left side;1
                    160;Rib 9 left side;1
                    161;Rib 10 left side;1
                    162;Rib 11 left side;1
                    163;Rib 12 left side;1
                    164;Rib 1 right side;1
                    165;Rib 2 right side;1
                    166;Rib 3 right side;1
                    167;Rib 4 right side;1
                    168;Rib 5 right side;1
                    169;Rib 6 right side;1
                    170;Rib 7 right side;1
                    171;Rib 8 right side;1
                    172;Rib 9 right side;1
                    173;Rib 10 right side;1
                    174;Rib 11 right side;1
                    175;Rib 12 right side;1
                    176;Scapulae;1
                    177;Skull;1
                    180;Cervical vertebra 1;1
                    181;Cervical vertebra 2;1
                    182;Cervical vertebra 3;1
                    183;Cervical vertebra 4;1
                    184;Cervical vertebra 5;1
                    185;Cervical vertebra 6;1
                    186;Cervical vertebra 7;1
                    188;Thoracic vertebra 1;1
                    189;Thoracic vertebra 2;1
                    190;Thoracic vertebra 3;1
                    191;Thoracic vertebra 4;1
                    192;Thoracic vertebra 5;1
                    193;Thoracic vertebra 6;1
                    194;Thoracic vertebra 7;1
                    195;Thoracic vertebra 8;1
                    196;Thoracic vertebra 9;1
                    197;Thoracic vertebra 10;1
                    198;Thoracic vertebra 11;1
                    199;Thoracic vertebra 12;1
                    201;Lumbar vertebra 1;1
                    202;Lumbar vertebra 2;1
                    203;Lumbar vertebra 3;1
                    204;Lumbar vertebra 4;1
                    205;Lumbar vertebra 5;1
                    206;Sacrum;1
                    207;Sternum;1
                    211;Head skin;2
                    212;Trunk skin;2
                    214;Left arm skin;2
                    215;Right arm skin;2
                    217;Left leg skin;2
                    218;Right leg skin;2
                    219;Small intestine (wall + contents);4
                    220;Spinal cord;4
                    221;Spleen;4
                    222;Stomach wall;4
                    223;Teeth;1
                    224;Testes;4
                    225;Thymus;4
                    226;Thyroid;4
                    227;Trachea;4
                    242;Bladder contents;4
                    243;Large intestine contents;4
                    244;Stomach contents;4
                    245;Bed + pillow;4
                    250;Ascending + transverse colon;4
                    251;Descending + sigmoid colon;4
                    252;Gall bladder;4
                    253;Prostate;4"""
def golem_organs():
    lg = __golem_organs.replace('\n', ';').replace('\r', '').split(';')

    while len(lg) > 0:
        org = lg.pop(0)
        desc = lg.pop(0)
        tiss = lg.pop(0)
        yield int(org), desc.strip(), __tissues[int(tiss)]



import numpy as np
from .CPF100 import CPF100_M, CPF100_F, CPF100_u, CPF100_M_px, CPF100_F_px
from . import elements
import scipy.interpolate as inter

###############################################################################
# Calculates an emmission spectra from Bhremsstralong on electron interaction
# in tungsten. I.E spectra from an x-ray tube with tungsten anode.
# This code is based on:
#
# PAPER 1
# Calculation of x-ray spectra emerging from an x-ray tube. Part I.
# Electron penetration characteristics in x-ray targets
# Gavin G. Poludniowskia and Philip M. Evans
# Joint Department of Physics, Institute of Cancer Research and
# The Royal Marsden NHS Foundation Trust,
# Downs Road, Sutton, Surrey SM2 5PT, United Kingdom
#
# and
#
# PAPER 2
# Calculation of x-ray spectra emerging from an x-ray tube. Part II.
# X-ray production and filtration in x-ray targets
# Gavin G. Poludniowskia
# Joint Department of Physics, Institute of Cancer Research and
# The Royal Marsden NHS Foundation Trust,
# Downs Road, Sutton, Surrey SM2 5PT, United Kingdom
###############################################################################

# some constants
TUNGSTEN_ATOMIC_NUMBER = 74
TUNGSTEN_DENSITY = 19.25  # g / cm3
AVOGADOS_NUMBER = 6.02214129e23
TUNGSTEN_MOLAR_MASS = 0.005439512619669277  # [mol / g]
TUNGSTEN_NUMBER_DENSITY = AVOGADOS_NUMBER * TUNGSTEN_ATOMIC_NUMBER * TUNGSTEN_MOLAR_MASS
ELECTRON_REST_MASS = 510.9989  # keV/c^2
CLASSIC_ELEC_RADII = 2.81794033E-13  # cm
FINE_STRUCTURE = 0.00729735257   # fine structure constant
THETA_BAR = (TUNGSTEN_ATOMIC_NUMBER * CLASSIC_ELEC_RADII)**2 * FINE_STRUCTURE



#interpolation of attinuation coeffisients obtained from XCOM NIST database
ATTINUATION_W = inter.interp1d(np.squeeze(elements.W_attinuation[0, :]),
                               np.squeeze(elements.W_attinuation[1, :]),
                               'linear',  bounds_error=False, fill_value=0)
ATTINUATION_Al = inter.interp1d(np.squeeze(elements.Al_attinuation[0, :]),
                                np.squeeze(elements.Al_attinuation[1, :]),
                                'linear',  bounds_error=False, fill_value=0)
ATTINUATION_Cu = inter.interp1d(np.squeeze(elements.Cu_attinuation[0, :]),
                                np.squeeze(elements.Cu_attinuation[1, :]),
                                'linear',  bounds_error=False, fill_value=0)
DENSITY = {'copper': 8.960,
           'tungsten': 19.25,
           'aluminum': 2.699}


def attinuation(hv, name='tungsten', density=False):
    """Attinuation (NIST data) for tungsten, aluminum and copper
        hv: photon energy
        name: material to calculate attinuation coefficient
        density: if True multiply attinuation coefficient by the
            material density
    """
    if name.lower() == 'tungsten':
        a = ATTINUATION_W
    elif name.lower() == 'aluminum':
        a = ATTINUATION_Al
    elif name.lower() == 'copper':
        a = ATTINUATION_Cu
    else:
        return np.zeros_like(hv)
    if density:
        return a(hv) * DENSITY[name]
    return a(hv)


def __R_TW(T0):
    """ Thomson-Whiddington electron range in tungsten.
        T0: Initial electron energy
    """
    return 0.0119 * T0**1.513


def __PDF_int_F(u, T0, px):
    """Forward probabilitydistribution of electrons in tungsten with energy
    u * T0 and depth px [mg/cm2]
    u: fraction of initial electron energy (tube potensial)
    T0: initial energy (tube potensial)
    px: depth in tungsten
    """
    # interpolating Monte-Carlo data obtained from ref1 (paper1)
    f = inter.interp2d(CPF100_F_px, CPF100_u, CPF100_F, fill_value=None)
    return f(px, u) * __R_TW(T0) / __R_TW(100.)


def __PDF_int_M(u, T0, px):
    """Backward (scattered) probability distribution of electrons in tungsten
    with energy u * T0 and depth px [mg/cm2]
    u: fraction of initial electron energy (tube potensial)
    T0: initial energy (tube potensial)
    px: depth in tungsten
    """
    # interpolating Monte-Carlo data obtained from ref1 (paper1)
    f = inter.interp2d(CPF100_M_px, CPF100_u, CPF100_M, fill_value=None)
    return f(px, u) * __R_TW(T0) / __R_TW(100.)


def __f(u, T0, px):
    """Combined probability of finding a electron at depth px with energy u*T0
    u: fraction of initial electron energy (tube potensial)
    T0: initial energy (tube potensial)
    px: depth in tungsten
    """
    # based directly on paper 1 eq 3
    rtw = 0.0119 * T0**1.513
    ppx = np.repeat(np.reshape(px, (1, -1)), u.shape[0], axis=0)

    n_f = (1. - ppx / rtw)**(1.753)
    exp_dum = 1. - np.exp(- 18. * ppx / rtw)
    B = 0.5 + .084 * exp_dum
    F = 0.584 * exp_dum
    n_m = n_f * B * (F+1.) / (1.-F*B)
    return n_f * __PDF_int_F(u, T0, px) + n_m * __PDF_int_M(u, T0, px)


def __cross_section_SRMEBH(hv, u, T0):
    """Modified  Bethe Heitler Cross section for emmission of bhremsstralung
    from electron->tungsten interaction.
    hv: photon energy emmitted
    u: fraction of initial energy for electron before interaction
    T0: Tube potensial(initial energy)
    """
    Ti = u * T0
    Ei = Ti + ELECTRON_REST_MASS
    Ef = Ei - hv
    pic = (Ei**2 - ELECTRON_REST_MASS**2)**.5
    pfc = (Ef**2 - ELECTRON_REST_MASS**2)**.5
    L = (2 * np.log((Ei * Ef + pic * pfc - ELECTRON_REST_MASS**2)
         / (ELECTRON_REST_MASS * hv)))
    a = THETA_BAR * 2*(4*Ei*Ef*L-7*pic*pfc) / (3*hv*pic*pic)*(pic/pfc)
    a[Ti <= (hv + 0.001)] = 0  # removing electrons with initial energy < hv
    return a


def __F(hv, px, theta):
    """Anode attinuation correction
    hv: photon energy [kev]
    px: emmission depth in tungsten [mg/cm2]
    theta: anode angle"""

    return np.exp(-attinuation(hv) * px / np.sin(theta) / 1000.)


def __cross_section_char(hv, T0):
    """ Cross section for characteristic radiation emmission
    hv: photon energy [ndarray]
    T0: tube potential
    """
    hvi = np.array([59.3, 58.0, 67.2, 69.1])
    fi = np.array([.505, .291, .162, .042])
    Carr = np.zeros_like(hv)

    if not any(hv >= hvi[0]):
        return Carr
    for i in range(4):
        ind = np.argmin(np.abs(hv-hvi[i]))
        Carr[ind] = (4.4*fi[i] * .94 / 2.) * __Nobserved_char(hvi[i], T0)
    return Carr


def __Nobserved_char(hv, T0):
    """Number of caracteristic photons emmitted for one electron interacting
        with one tungsten atom, se paper 2
    """
    u = np.linspace(.1, 1, 200).astype(np.double)
    px = np.linspace(0, 16, 100).astype(np.double)
    du = u[1] - u[0]
    dpx = (px[1] - px[0]) / 1000  # converting from mg/cm to g/cm
    a = np.repeat(__cross_section_SRMEBH(hv, u, T0).reshape((-1, 1)),
                  px.shape[0], axis=-1)
    corr_a = np.nan_to_num(a * __f(u, T0, px)).sum(axis=0)
    return np.nan_to_num(corr_a).sum()*du*dpx


def __Nobserved_emit(hv, T0, theta):
    """Number of bremssthralung photons emmitted for one electron interacting
        with one tungsten atom, se paper 2
    """
    u = np.linspace(.1, 1, 200).astype(np.double)
    px = np.linspace(0, 16, 100).astype(np.double)
    du = u[1] - u[0]
    dpx = (px[1] - px[0]) / 1000  # converting from mg/cm to g/cm
    a = np.repeat(__cross_section_SRMEBH(hv, u, T0).reshape((-1, 1)),
                  px.shape[0], axis=-1)

    corr_a = np.nan_to_num(a * __f(u, T0, px)).sum(axis=0) * __F(hv, px, theta)

    return np.nan_to_num(corr_a).sum() * du * dpx


def __raw_specter(T0, angle_deg=None, angle_rad=None):
    """Unfiltered specter from anode per electron
    T0: tube potensial
    angle_deg/angle_rad: anode angle in degrees or radians"""

    if angle_deg is not None:
        angle = np.deg2rad(angle_deg)
    elif angle_rad is not None:
        angle = angle_rad
    else:
        angle = np.deg2rad(12.)

    hv = np.linspace(T0 * .1, T0, 200).astype(np.double)
    Nemit = np.empty_like(hv, dtype=np.double)
    Nchar = __cross_section_char(hv, T0)
    for i, e in enumerate(hv):
        Nemit[i] = __Nobserved_emit(e, T0, angle)

    # n is the NA (Avogados constant) * tungsten molar number per gram
    # / by tube potential / tungsten density
#    n = 6.02214129e23*0.005439512619669277  / T0 / 19.3   *10
#    print n
    n = TUNGSTEN_NUMBER_DENSITY
#    print n
    dx = 2.
    Nf = 0.68
    P = 0.33 * 100
    return hv, (Nemit + Nchar * (1 + P)) * n * dx * Nf


def specter(T0, angle_deg=None, angle_rad=None,
            filtration_materials=None, filtration_mm=None,
            mAs=1., sdd=100., detector_area=None):
    """
    Generate x-ray specter from tungsten anode by the methods of ref. 1 and 2.

    INPUT:
        T0 : float
            Tube potential in keV should be between 30 and 150 keV but may
            provide acceptable results for higher potentials.
        angle_deg : float [degrees]
            Anode angle in degrees, typicalvalue is 12 degrees.
        angle_rad : float [radians]
            Anode angle in radians.
        filtration_materials : iterable or string
            Additional Aluminum or Copper filtration in the tube
            Iterable may contain the following values 'al', 'aluminum' or 13
            for aluminum and similar for copper. One of the values may also be
            supplied as a string instead of iterator.
        filtration_mm : iterable or float
            mm of Copper or Aluminum filtration, must be provide with the
            filtration_materials keyword and of same type.
        mAs : float
            Scaling of specter to the mAs setting of the tube.
        sdd : float > 0.0
            Source detector distance in cm, or simply the distance to calculate
            the spectrum intensities.
        detector_area : float
            Area to calculate spectrum intensities at sdd.
    OUTPUT:
        energies, intensity : {float ndarray, float ndarray}
            The spectrum energies and intensity, energy in units of electron
            volts eV, intensity are gives as N_photons / [cm^2 mAs]

    Reference 1:
        Calculation of x-ray spectra emerging from an x-ray tube. Part I.
        Electron penetration characteristics in x-ray targets.
        Medical Physics, Vol. 34(6), 2164-2174

        Gavin G. Poludniowskia and Philip M. Evans
        Joint Department of Physics, Institute of Cancer Research and
        The Royal Marsden NHS Foundation Trust,
        Downs Road, Sutton, Surrey SM2 5PT, United Kingdom

    Reference 2:
        Calculation of x-ray spectra emerging from an x-ray tube. Part II.
        X-ray production and filtration in x-ray targets.
        Medical Physics, Vol. 34(6), 2175-2186

        Gavin G. Poludniowskia
        Joint Department of Physics, Institute of Cancer Research and
        The Royal Marsden NHS Foundation Trust,
        Downs Road, Sutton, Surrey SM2 5PT, United Kingdom
    """
    hv, N_obs = __raw_specter(T0, angle_deg, angle_rad)
    if isinstance(filtration_materials, str):
        filtration_materials = [filtration_materials]
    if not hasattr(filtration_mm, '__iter__'):
        filtration_mm = [filtration_mm]
    try:
        for material, mm in zip(filtration_materials, filtration_mm):
            if str(material).strip().lower() in ['al', 'aluminum', '13']:
                name = 'aluminum'
            elif str(material).strip().lower() in ['cu', 'copper', '29']:
                name = 'copper'
            elif str(material).strip().lower() in ['w', 'tungsten', '74',
                                                   'wolfram']:
                name = 'tungsten'
            else:
                continue
            N_obs *= np.exp(-attinuation(hv, name, True) * mm / 10.)
    except Exception as e:
        raise e
    electrons_per_mas = 1. / 1.60217657e-16
    if detector_area is None:
        red_factor = 1.
    else:
        red_factor = detector_area / (4 * np.pi * sdd**2)
    return (hv * 1000, N_obs * mAs * electrons_per_mas * red_factor)



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 28 14:30:02 2015

@author: erlean
"""
import numpy as np
from scipy.ndimage.interpolation import affine_transform, spline_filter
from scipy.ndimage.filters import gaussian_filter
import itertools
from opendxmc.engine import Engine

from opendxmc.tube.tungsten import specter as tungsten_specter
from opendxmc.runner.ct_sources import ct_source_space
from opendxmc.runner.ct_sources import ct_seq
from opendxmc.utils import circle_mask
import time
from opendxmc.utils import human_time, rebin

import logging
logger = logging.getLogger('OpenDXMC')


def log_elapsed_time(time_start, elapsed_exposures, total_exposures,
                     start_exposure, n_histories=None):
    time_delta = time.clock() - time_start
    p = np.round(float(elapsed_exposures) / float(total_exposures) * 100., 1)
    eta = time_delta * (float(total_exposures-start_exposure) / float(elapsed_exposures-start_exposure) - 1.)
    if elapsed_exposures == total_exposures:
        if n_histories is not None:
            logger.info('{0}: [{1}%] Finished {2} histories in {3}'.format(
                time.ctime(), p, n_histories*total_exposures,
                human_time(time_delta)))
        else:
            logger.info('{0}: [{1}%] Finished {2} exposures in {3}'.format(
                time.ctime(), p, total_exposures, human_time(time_delta)))
    else:
        logger.info('{0}: [{1}%] estimated ETA is {2}, finished exposure {3}'
                    ' of {4}'.format(time.ctime(), p, human_time(eta),
                                     elapsed_exposures, total_exposures))
    return '{0}% ETA: '.format(p) + human_time(eta)


def recarray_to_dict(arr, key=None, value=None, value_is_string=False):
    key, value = arr.dtype.names
    assert key in arr.dtype.names
    assert value in arr.dtype.names
    if value_is_string:
        return {arr[key][i]: str(arr[value][i], encoding='utf-8')
                for i in range(arr.shape[0])}
    return {arr[key][i]: arr[value][i] for i in range(arr.shape[0])}


def generate_attinuation_lut(materials, material_map, min_eV=None,
                             max_eV=None, ignore_air=False):

    if isinstance(material_map, np.recarray):
        material_map = recarray_to_dict(material_map, value_is_string=True)
    elif isinstance(material_map, np.ndarray):
        material_map = recarray_to_dict(material_map, value_is_string=True)
    assert isinstance(material_map, dict)
    if min_eV is None:
        min_eV = 0.
    if max_eV is None:
        max_eV = 500.e3

    names = [m.name for m in materials]
    atts = {}
    for key, value in list(material_map.items()):
        key = int(key)
        try:
            ind = names.index(value)
        except ValueError:
            raise ValueError('No material named '
                             '{0} in first argument. '
                             'The material_map requires {0} material'.format(value))
        atts[key] = materials[ind].attinuation
        if value == 'air':
            air_key = key

    energies = np.unique(np.hstack([a['energy'] for a in list(atts.values())]))
    e_ind = (energies <= max_eV) * (energies >= min_eV)
    if not any(e_ind):
        raise ValueError('Supplied minimum or maximum energies '
                         'are out of range')
    energies = energies[e_ind]
    lut = np.empty((len(atts), 5, len(energies)), dtype='float64')
    for i, a in list(atts.items()):
        lut[i, 0, :] = energies
        if ignore_air and air_key == i:
            lut[i, 1:, :] = 0
        else:
            for j, key in enumerate(['total', 'rayleigh', 'photoelectric',
                                     'compton']):
                lut[i, j+1, :] = np.interp(energies, a['energy'], a[key])
    return lut

def prepare_geometry_from_organ_array(organ, organ_material_map, scale, materials):
        """genereate material and density arrays and material map from
           a list of materials to use
           INPUT:
               specter for this study
               list of materials
           OUTPUT :
               material_map, material_array, density_array
        """
        if organ is None:
            raise ValueError('No organ definitions in phantom')

        organ_material_map = recarray_to_dict(organ_material_map,
                                              value_is_string=True)
        #testing if we have the materials we need
        material_names = {m.name: m for m in materials}
        for value in organ_material_map.values():
            if not value in material_names:
                raise ValueError('Phantom requires missing material {}'.format(value))

#        organ = affine_transform(organ,
#                                 scale,
#                                 output_shape=np.floor(np.array(organ.shape)/scale),
#                                 cval=0, output=np.uint8, prefilter=True,
#                                 order=0).astype(np.uint8)
        scale_int = list([int(s) for s in scale])
        organ = organ[::scale_int[0], ::scale_int[1], ::scale_int[2]]
        material_array = np.asarray(organ, dtype=np.uint8)
        density_array = np.zeros(organ.shape, dtype='float64')

        material_map = {}
        key = 0
        for i in np.unique(organ):
            try:
                material_name = organ_material_map[i]
            except:
                import pdb
                pdb.set_trace()
            if material_name not in material_map:
                material_map[material_name] = key
                key += 1

            ind = np.nonzero(organ == i)
            density_array[ind] = material_names[material_name].density
            material_array[ind] = material_map[material_name]


        # reversing material_map
        material_map = {key: value for value, key in material_map.items()}
        return material_map, material_array.astype('int32'), density_array

def attinuation_to_ct_numbers(material_attinuations, air_key, water_key):
    a = 1000./(material_attinuations[water_key] -material_attinuations[air_key])
    b = -a * material_attinuations[water_key]
    atts = []

    for key, att in material_attinuations.items():
        HU = a * att + b
        if HU > 1000:
            HU=1000
        atts.append((key,HU))
    return atts

def prepare_geometry_from_ct_array(ctarray, scale ,specter, materials):
        """genereate material and density arrays and material map from
           a list of materials to use
           INPUT:
               specter for this study
               list of materials
           OUTPUT :
               material_map, material_array, density_array
        """
        if ctarray is None:
            return
        ctarray = gaussian_filter(ctarray, scale)
#        ctarray = affine_transform(spline_filter(ctarray,
#                                                 order=3,
#                                                 output=np.int16),
#                                   scale,
#                                   output_shape=np.floor(np.array(ctarray.shape)/scale),
#                                   cval=-1000, output=np.int16, prefilter=False)
        ctarray = rebin(ctarray, scale)
        specter = (specter[0], specter[1]/specter[1].mean())

        water_key = None
        material_map = {}
        material_att = {}
        material_dens = {}
        materials = [m for m in materials if m.organic]
        materials.sort(key=lambda x: x.density)
        for i, mat in enumerate(materials):
            material_map[i] = mat.name
            material_dens[i] = float(mat.density)
            # interpolationg and integrating attinuation coefficient
#            material_att[i] =
#            material_att[i] = np.trapz(np.interp(specter[0],
#                                                 mat.attinuation['energy'],
#                                                 mat.attinuation['total']),
#                                       specter[0])
            material_att[i] = np.sum(np.exp(-np.interp(specter[0],
                                     mat.attinuation['energy'],
                                     mat.attinuation['total']))*specter[1]
                           )

            material_att[i] *= mat.density

            if mat.name == 'water':
                water_key = i
            elif mat.name == 'air':
                air_key = i

        assert water_key is not None  # we need to include water in materials
        assert air_key is not None  # we need to include water in materials

        # getting a list of attinuation
        material_HU_list = [(key, (att / material_att[water_key] - 1.)*1000.)
                            for key, att in material_att.items()]
        material_HU_list = attinuation_to_ct_numbers(material_att, air_key, water_key)
        material_HU_list.sort(key=lambda x: x[1])
        HU_bins = (np.array(material_HU_list)[:-1, 1] +
                   np.array(material_HU_list)[1:, 1]) / 2.

        if HU_bins[-2] < 300:
            HU_bins[-1] = 300

        material_array = np.digitize(
            ctarray.ravel(), HU_bins).reshape(ctarray.shape).astype('int32')
        # Using densities fromdefined materials
        density_array = np.asarray(material_array, dtype='float64')
        np.choose(material_array,
                  [material_dens[i] for i, _ in material_HU_list],
                  out=density_array)

        # using densities defined by HU units in ct array
        HU_list = np.zeros(len(material_HU_list), dtype=np.int)
        for ind, HU in material_HU_list:
            HU_list[ind] = HU
        ctarray_model = np.choose(material_array, HU_list)
        density_array *= np.clip((ctarray - ctarray.min()+1) / (ctarray_model - ctarray.min() + 1), 0.5, 1.5)
        ##########################

        return material_map, material_array, density_array


def ct_runner_validate_simulation(materials, simulation, ctarray=None, organ=None,
                                  organ_material_map=None):
    """
    validating a ct mc simulation
    """

    # testing for required attributes
    if ctarray is not None:
        logger.info('Recalculating material mapping from CT array for {}'.format(simulation['name']))
        specter = tungsten_specter(simulation['aquired_kV'],
                                   filtration_materials='al',
                                   filtration_mm=simulation['al_filtration'])
        vals = prepare_geometry_from_ct_array(ctarray,
                                              simulation['scaling'],
                                              specter,
                                              materials)
    elif (organ is not None) and (organ_material_map is not None):
        logger.info('Recalculating material mapping from organ array for {}'.format(simulation['name']))

        vals = prepare_geometry_from_organ_array(organ,
                                                 organ_material_map,
                                                 simulation['scaling'],
                                                 materials)
    else:
        logger.warning('CT study {} has no CT images or organ arrays. Simulation not '
                       'started'.format(simulation['name']))
        raise ValueError('CT study {} must have CT images or organ array to run a '
                         'simulation'.format(simulation['name']))


    material_map = vals[0]
    material = vals[1]
    density = vals[2]

    # Testing for required materials if the simulation have a material_map
    material_names = [m.name for m in materials]
    for m_name in material_map.values():
        if m_name not in material_names:
            raise ValueError('Provided materials are not in ct study')

    # test for correct material geometry and mapping
    materials_ind = list(np.unique(material))
    materials_key = list(material_map.keys())
    materials_range = list(range(len(materials_ind)))

    try:
        for ind, key in zip(materials_ind, materials_key):
            assert key == ind
            assert key in materials_range

    except AssertionError:
        logger.warning('Something went wrong with material definitions, '
                       'attempting to recreate material mapping for '
                       'current simulation.')
        material_map = None
        raise ValueError('Error in material definitions for simulation')
    return material.astype('int32'), material_map, density.astype('float64')


def ct_runner(materials, simulation, ctarray=None, organ=None,
              organ_material_map=None, exposure_modulation=None,
              energy_imparted_to_dose_conversion=True, callback=None,
              energy_imparted=None, **kwargs):
    """Runs a MC simulation on a simulation object, and updates the
    energy_imparted property.

    INPUT:

        simulation :

        materials : a list of Material instances to be used in the simulation

        ignore_air : [optional] If set, ignores the material 'air' in
            MC calculation

    OUTPUT:
        None, but updates the energy_imparted property of simulation
    """
    logger.info('Preparing simulation for {}'.format(simulation['name']))
    if simulation['is_phantom']:
        materials_organic = [m for m in materials]
    else:
        materials_organic = [m for m in materials if m.organic]

    # Validating if everything is in place
    material, material_map, density = ct_runner_validate_simulation(materials, simulation,
                                                                    ctarray=ctarray,
                                                                    organ_material_map=organ_material_map,
                                                                    organ=organ)

    phase_space = ct_source_space(simulation, exposure_modulation)

    N = np.array(material.shape, dtype='int32')

    offset = np.zeros(3, dtype='float64')
    spacing = (simulation['spacing'] * simulation['scaling']).astype('float64')

    lut = generate_attinuation_lut(materials_organic, material_map,
                                   max_eV=500.e3,
                                   ignore_air=simulation['ignore_air'])
    del energy_imparted
    energy_imparted = None
    if energy_imparted is None:
        energy_imparted = np.zeros_like(density, dtype='float64')


    tot_histories = simulation['histories'] * simulation['exposures']

    if tot_histories > 1e7:
        coffe_msg = ', go get coffe!'
    else:
        coffe_msg = '.'
    logger.info('{0}: Starting simulation with about {1} histories per '
                'rotation{2}'.format(time.ctime(), tot_histories, coffe_msg))

    lut_shape = np.array(lut.shape, dtype='int32')

    logger.info('Initializing geometry')
    use_siddon = np.array([simulation['use_siddon']], dtype='int32')
    engine = Engine()
    geometry = engine.setup_simulation(N, spacing, offset, material, density, lut_shape, lut, energy_imparted, use_siddon)

    start_exposure = simulation['start_at_exposure_no']
    time_start = time.clock()
    exposure_time = time_start

    for p, e, n in phase_space:
        source = engine.setup_source_bowtie(*p)
        engine.run_bowtie(source, simulation['histories'], geometry)
        engine.cleanup(source=source)

        if (time.clock() - exposure_time) > 5:
            eta = log_elapsed_time(time_start, e+1, n, start_exposure)
            if callback is not None:
                callback(simulation['name'], progressbar_data=[np.squeeze(energy_imparted.max(axis=0)), spacing[1] ,spacing[2] ,eta, True])
#                simulation['start_at_exposure_no'] = e + 1
            exposure_time = time.clock()


    engine.cleanup(simulation=geometry)
#    time_start = time.clock()
#    for p, e, n in phase_space:
#        score_energy(p, N, spacing, offset, material,
#                     density, lut, energy_imparted)
#        eta = log_elapsed_time(time_start, e+1, n, start_exposure)
#        if callback is not None:
#            callback(simulation['name'], {'energy_imparted': energy_imparted}, e + 1, eta)
#        simulation['start_at_exposure_no'] = e + 1

    if callback is not None:
        callback(simulation['name'], progressbar_data=[np.squeeze(energy_imparted.max(axis=0)), spacing[1] ,spacing[2] ,'Preforming dose calibration', True])
#        callback(simulation['name'], {'energy_imparted': None}, e + 1, 'Preforming dose calibration')

    generate_dose_conversion_factor(simulation, materials, callback)

    if callback is not None:
        callback(simulation['name'], {'energy_imparted': None}, e + 1, progressbar_data=[np.squeeze(energy_imparted.max(axis=1)), spacing[0] ,spacing[2] ,'Done', False])
    simulation['start_at_exposure_no'] = 0
    simulation['MC_finished'] = True
    simulation['MC_running'] = False
    simulation['MC_ready'] = False
    return simulation, {'density': density, 'material': material, 'material_map': material_map, 'energy_imparted': energy_imparted}


def generate_dose_conversion_factor(simulation, materials, callback=None):
    air, pmma = None, None
    for m in materials:
        if m.name == 'pmma':
            pmma = m
        elif m.name == 'air':
            air = m

    if (simulation['ctdi_air100'] > 0.) and (air is not None):
        obtain_ctdiair_conversion_factor(simulation, air, callback=callback)
    if (simulation['ctdi_w100'] > 0.) and (pmma is not None) and (air is not None):
        size = simulation['ctdi_phantom_diameter']
        obtain_ctdiw_conversion_factor(simulation, pmma, air, size=size, callback=callback)
#    else:
#        msg = """Need a combination of air material and ctdi air or ctdi_w100
#                 pmma material and ctdiw_100 to generate energy to dose
#                 conversion factor."""
#        logger.warning(msg)



def obtain_ctdiair_conversion_factor(simulation, air_material, callback=None):

    logger.info('Starting simulating CTDIair100 measurement for '
                '{0}. CTDIair100 is {1}mGy'.format(simulation['name'], simulation['ctdi_air100']))
    spacing = np.array((2, 2, 10), dtype='float64')

    N = np.rint(np.array((simulation['sdd'] / spacing[0],
                          simulation['sdd'] / spacing[1], 3),
                         dtype='int32')).astype('int')

    offset = (-N * spacing / 2.).astype('float64')
    material_array = np.zeros(N, dtype='int32')
    material_map = {0: air_material.name}
    density_array = np.zeros(N, dtype='float64') + air_material.density

    lut = generate_attinuation_lut([air_material], material_map, max_eV=0.15e6).astype('float64')
    lut_shape = np.array(lut.shape, dtype='int32')
    dose = np.zeros_like(density_array, dtype='float64')

    
    total_collimation = simulation['detector_rows'] * simulation['detector_width']


    engine = Engine()
    use_siddon = np.array([simulation['use_siddon']], dtype=np.int)
    geometry = engine.setup_simulation(N, spacing, offset, material_array, density_array, lut_shape, lut, dose, use_siddon)
    teller = 0
    center = np.floor(N / 2).astype(np.int)
    center_dose = 0
    t1 = time.clock()
    while center_dose < 150000:
        if teller > 0:
            logger.debug('Not sufficient data, running again. Dose in center is now {0}, max dose: {1}.'.format(center_dose, dose.max()))
        teller += 1
        
        if simulation['use_tube_B']:
            wA = simulation.get('tube_weight_A')
            wB = simulation.get('tube_weight_A')
            wS = wA + wB
            en_specter_A = tungsten_specter(simulation['kV_A'], 
                              angle_deg=simulation['anode_angle'],
                              filtration_materials='Al',
                              filtration_mm=simulation['al_filtration'])
        
            phase_space_A = ct_seq(simulation['scan_fov'], simulation['sdd'],
                                 total_collimation, start=0, stop=0, step=1,
                                 exposures=simulation['exposures'],
                                 histories=simulation['histories'],
                                 energy_specter=en_specter_A,
                                 bowtie_radius=simulation['bowtie_radius'],
                                 bowtie_distance=simulation['bowtie_distance'],
                                 weight=wA/wS                
                                 )
            en_specter_B = tungsten_specter(simulation['kV_B'], 
                              angle_deg=simulation['anode_angle'],
                              filtration_materials='Al',
                              filtration_mm=simulation['al_filtration'])
        
            phase_space_B = ct_seq(simulation['scan_fov'], simulation['sdd'],
                                 total_collimation, start=0, stop=0, step=1,
                                 exposures=simulation['exposures'],
                                 histories=simulation['histories'],
                                 energy_specter=en_specter_B,
                                 bowtie_radius=simulation['bowtie_radius'],
                                 bowtie_distance=simulation['bowtie_distance'],
                                 weight=wB/wS                
                                 )
            phase_space = itertools.chain(phase_space_A, phase_space_B)
            
        else:
    
            en_specter = tungsten_specter(simulation['kV_A'], 
                                          angle_deg=simulation['anode_angle'],
                                          filtration_materials='Al',
                                          filtration_mm=simulation['al_filtration'])
            
            phase_space = ct_seq(simulation['scan_fov'], simulation['sdd'],
                                 total_collimation, start=0, stop=0, step=1,
                                 exposures=simulation['exposures'],
                                 histories=simulation['histories']/2,
                                 energy_specter=en_specter,
                                 bowtie_radius=simulation['bowtie_radius'],
                                 bowtie_distance=simulation['bowtie_distance']
                                 )


        for batch, e, n in phase_space:
            source = engine.setup_source_bowtie(*batch)
            engine.run_bowtie(source, simulation['histories'], geometry)
            engine.cleanup(source=source)
    #        break
            if (time.clock() - t1) > 1:
#                eta = log_elapsed_time(t0, e+1, n, 0)
                t1 = time.clock()
                if callback:
                    callback(simulation['name'], progressbar_data=[np.squeeze(dose.max(axis=2)), spacing[0] ,spacing[1] ,'Run number {0}'.format(teller+1), True])
#                    callback(simulation['name'], {'energy_imparted':dose}, 0, '', save=False)
        center_dose += np.sum(dose[center[0], center[1], center[2]])
    callback(simulation['name'], progressbar_data=[np.squeeze(dose.max(axis=2)), spacing[0] ,spacing[1] ,'Done', True])
    engine.cleanup(simulation=geometry)

#    engine.cleanup(simulation=geometry, energy_imparted=dose)
#    dose = gaussian_filter(dose, (1., 1., 0.))

    d = center_dose / teller / (air_material.density * np.prod(spacing)) / simulation['pitch']
    simulation['conversion_factor_ctdiair'] = np.nan_to_num(1. / d * total_collimation)
    return dose

def generate_ctdi_phantom(simulation, pmma, air, size=32., callback=None):
    spacing = np.array((.1, .1, 2.5), dtype='float64')
#    N = np.rint(np.array((simulation['sdd'] / spacing[0],
#                          simulation['sdd'] / spacing[0], 6),
#                         )).astype('int32')
    N = np.rint(np.array((size*1.15 / spacing[0],
                          size*1.15 / spacing[0], 6),
                         )).astype('int32')

    offset = (-N * spacing / 2.).astype('float64')
    material_array = np.zeros(N, dtype='int32')
    radii_phantom = size / spacing[0] / 2
    radii_meas = (1.3 / 2 / spacing[0])
    center = (N / 2.)[:2]
    radii_pos = radii_phantom - (1 + 1.3 / 2) / spacing[0]
    pos = [(center[0], center[1])]

    for ang in [0, 90, 180, 270]:
        dx = radii_pos * np.sin(np.deg2rad(ang))
        dy = radii_pos * np.cos(np.deg2rad(ang))
        pos.append((center[0] + dx, center[1] + dy))

    measure_indices = []

    for i in range(int(N[2])):
        material_array[:, :, i] += circle_mask((N[0], N[1]),
                                               radii_phantom)
#        for x, y in pos:
#            material_array
        for p in pos:
            mask = circle_mask((N[0], N[1]), radii_meas, center=p)
            material_array[:, :, i] += mask
            if i == 0:
                measure_indices.append(np.argwhere(mask))

    material_map = {0: air.name, 1: pmma.name, 2: air.name}
    density_array = np.zeros_like(material_array, dtype='float64')
    density_array[material_array == 0] = air.density
    density_array[material_array == 1] = pmma.density
    density_array[material_array == 2] = air.density

    lut = generate_attinuation_lut([air, pmma], material_map)
    return N, spacing, offset, material_array, density_array, lut, measure_indices


def obtain_ctdiw_conversion_factor(simulation, pmma, air,
                                   size=32., callback=None):

    logger.info('Starting simulating CTDIw100 measurement for '
                '{}'.format(simulation['name']))
    args = generate_ctdi_phantom(simulation, pmma, air, size=size)
    N, spacing, offset, material_array, density_array, lut, meas_pos = args

    lut_shape = np.array(lut.shape, dtype='int32')

    dose = np.zeros_like(density_array, dtype='float64')

    total_collimation = simulation['detector_rows'] * simulation['detector_width']


    
    
    if simulation['use_tube_B']:
        wA = simulation.get('tube_weight_A')
        wB = simulation.get('tube_weight_A')
        wS = wA + wB
        en_specter_A = tungsten_specter(simulation['kV_A'], 
                          angle_deg=simulation['anode_angle'],
                          filtration_materials='Al',
                          filtration_mm=simulation['al_filtration'])
    
        phase_space_A = ct_seq(simulation['scan_fov'], simulation['sdd'],
                             total_collimation, start=0, stop=0, step=1,
                             exposures=simulation['exposures'],
                             histories=simulation['histories'],
                             energy_specter=en_specter_A,
                             bowtie_radius=simulation['bowtie_radius'],
                             bowtie_distance=simulation['bowtie_distance'],
                             weight=wA/wS,
                             xcare=simulation['xcare']
                             )
        en_specter_B = tungsten_specter(simulation['kV_B'], 
                          angle_deg=simulation['anode_angle'],
                          filtration_materials='Al',
                          filtration_mm=simulation['al_filtration'])
    
        phase_space_B = ct_seq(simulation['scan_fov'], simulation['sdd'],
                             total_collimation, start=0, stop=0, step=1,
                             exposures=simulation['exposures'],
                             histories=simulation['histories'],
                             energy_specter=en_specter_B,
                             bowtie_radius=simulation['bowtie_radius'],
                             bowtie_distance=simulation['bowtie_distance'],
                             weight=wB/wS,
                             xcare=simulation['xcare']
                             )
        phase_space = itertools.chain(phase_space_A, phase_space_B)
        
    else:

        en_specter = tungsten_specter(simulation['kV_A'], 
                                      angle_deg=simulation['anode_angle'],
                                      filtration_materials='Al',
                                      filtration_mm=simulation['al_filtration'])
        
        phase_space = ct_seq(simulation['scan_fov'], simulation['sdd'],
                             total_collimation, start=0, stop=0, step=1,
                             exposures=simulation['exposures'],
                             histories=simulation['histories']/2,
                             energy_specter=en_specter,
                             bowtie_radius=simulation['bowtie_radius'],
                             bowtie_distance=simulation['bowtie_distance'],
                             xcare=simulation['xcare'])

    
                         
                         
    use_siddon = np.array([simulation['use_siddon']], dtype=np.int)
    engine = Engine()
    geometry = engine.setup_simulation(N, spacing, offset, material_array,
                            density_array, lut_shape, lut, dose, use_siddon)


    history_factor = int(1e8 / simulation['histories'] / simulation['exposures'])
    if history_factor < 1:
        history_factor = 1
    histories = simulation['histories'] * history_factor
    t0 = time.clock()
    t1 = t0


    for batch, e, n in phase_space:
        source = engine.setup_source_bowtie(*batch)
        engine.run_bowtie(source, histories, geometry)
        engine.cleanup(source=source)

        if (time.clock() - t1) > 5:
            eta = log_elapsed_time(t0, e+1, n, 0)
            if callback:
                callback(simulation['name'], progressbar_data=[np.squeeze(dose.max(axis=2)), spacing[0] ,spacing[1] ,eta, True])
#                callback(simulation['name'], {'energy_imparted':dose}, 0, '', save=False)
            t1 = time.clock()

    if callback:
        callback(simulation['name'], progressbar_data=[np.squeeze(dose.max(axis=2)), spacing[0] ,spacing[1] ,'Done', True])
#        callback(simulation['name'], {'energy_imparted':dose}, 0, 'Done', save=False)

    dose /= history_factor

    engine.cleanup(simulation=geometry)
#
#    plt.imshow(dose[:,:,1])
#    plt.show()
#    dose = gaussian_filter(dose, [.5, .5, .1])
#    plt.imshow(dose[:,:,1])
#    plt.show()
    d = []
    for p in meas_pos:
        x, y = p[:, 0], p[:, 1]
        d.append(dose[x, y, 1:-2].mean() / (air.density * np.prod(spacing)))
        dose[x, y, 1:-2] = dose.max()
    logger.debug('Estimated dose for CTDI phantom; center: {0}, periphery: {1}, {2}, {3}, {4}.'.format(*d))
#    pdb.set_trace()
    ctdiv = d.pop(0) / 3.
    ctdiv += (2. * sum(d) / 3. / 4.)
    simulation['conversion_factor_ctdiw'] = np.nan_to_num(simulation['ctdi_w100'] / ctdiv * total_collimation)
#    pdb.set_trace()
#    import pdb
#    pdb.set_trace()
    return dose



# -*- coding: utf-8 -*-
"""
Created on Tue Sep  8 10:10:58 2015

@author: erlean
"""
import numpy as np
from PyQt5 import QtGui, QtCore, QtWidgets
from scipy.ndimage.interpolation import affine_transform
from opendxmc.database import Database, PROPETIES_DICT_TEMPLATE, Validator, PROPETIES_DICT_TEMPLATE_GROUPING
from opendxmc.database.import_phantoms import read_phantoms
from opendxmc.database import import_ct_series
from opendxmc.runner import ct_runner
from opendxmc.utils import find_all_files
import logging
logger = logging.getLogger('OpenDXMC')


class ImportScalingValidator(QtGui.QValidator):
    def __init__(self, parent):
        super().__init__(parent)

    def fixup(self, instr):
        instr = ''.join([b for b in instr.replace(',', ' ') if b in "1234567890. "])

        fstr = ""
        d_in_word = False
        for s in instr:
            if s == '.':
                if not d_in_word:
                    fstr += s
                else:
                    fstr += ' '
                d_in_word = True
            elif s == ' ':
                d_in_word = False
                fstr += ' '
            else:
                fstr += s


        return fstr
#        nums = [float(a) for a in instr.split()]
#        return ' '.join([str(n) for n in nums])

    def validate(self, rawstr, pos):
        instr = ''.join([b for b in rawstr.replace(',', ' ') if b in "1234567890. "])
        pos -= (len(instr) - len(rawstr))

        if len(instr) < 5:
            return self.Intermediate, instr, pos
        numbers = []
        state = self.Acceptable
        last = instr[-1]
        for word in instr.split():
            try:
                float(word)
            except:
                state = self.Intermediate
            numbers.append(word)
        if len(numbers) > 3:
            rstr = ' '.join(numbers[:3])
        else:
            rstr = ' '.join(numbers)
        if last == ' ':
            rstr += last
        return state, rstr, pos


class ImportScalingEdit(QtWidgets.QLineEdit):
    request_set_import_scaling = QtCore.pyqtSignal(tuple)
    def __init__(self, importer, parent=None):
        super().__init__(parent)
        self.base_color = self.palette().color(self.palette().Base)

        self.request_set_import_scaling.connect(importer.set_import_scaling)
        help_txt = "Select scaling factor per dimension when importing DICOM images"
        self.setWhatsThis(help_txt)
        self.setToolTip(help_txt)
        self.editingFinished.connect(self.set_import_scaling)
        self.textEdited.connect(self.text_was_edited)
        self.setValidator(ImportScalingValidator(self))
        self.setText("2.0 2.0 1.0")
        self.set_import_scaling()

    @QtCore.pyqtSlot(str)
    def text_was_edited(self, txt):
        palette = self.palette()
        palette.setColor(palette.Base, QtCore.Qt.red)
        self.setPalette(palette)

    @QtCore.pyqtSlot()
    def set_import_scaling(self):
        txt = self.text()
        d = tuple(float(s) for s in txt.split())
        self.request_set_import_scaling.emit(d)
        self.setText(' '.join([str(n) for n in d]))
        palette = self.palette()
        palette.setColor(palette.Base, self.base_color)
        self.setPalette(palette)

class ArrayBuffer(object):
    def __init__(self):
        self.array_name = ""
        self.name = ""
        self.buffer_size = 2  # number of slices to buffer
        self.indices = np.arange(2, dtype=np.int)
        self.slices = np.empty((2,2, 2))
        self.view_orientation = 2
        self.index_threshold_edge = (0, 1)

    def set_buffer(self, slices, name, array_name, indices, orientation):
        self.indices = indices
        self.slices = slices
        self.name = name
        self.array_name = array_name
        self.view_orientation = orientation

        self.buffer_size = slices.shape[orientation]
        self.index_threshold_edge = (indices[0], indices[-1])
        logger.debug('Updating buffer, indices: {}'.format(indices))


    def is_slice_available(self, name, array_name, index, orientation):
        if orientation != self.view_orientation:
            return False, 0
        if name != self.name:
            return False, 0
        if array_name != self.array_name:
            return False, 0
        if index in self.indices:
            if index == self.index_threshold_edge[0]:
                return True, -1
            if index == self.index_threshold_edge[1]:
                return True, 1
            return True, 0
        return False, 0

    def get_slice(self, name, array_name, index, orientation):
        if orientation != self.view_orientation:
            raise ValueError('Not same orientation')
        if name != self.name:
            raise ValueError('Not same name')
        if array_name != self.array_name:
            raise ValueError('Not same array_name')
        if index not in self.indices:
            raise ValueError('Array not in buffer')
        sub_ind = np.nonzero(self.indices == index)
        if self.view_orientation == 0:
            return np.squeeze(self.slices[sub_ind, :, :])
        if self.view_orientation == 1:
            return np.squeeze(self.slices[:, sub_ind, :])
        return np.squeeze(self.slices[:, :, sub_ind])


class DatabaseInterface(QtCore.QObject):
    """ Async database interface, async provided with signal/slots resulting in
    two connections per task wished to be done, ie, signal/slot hell
    """
    database_busy = QtCore.pyqtSignal(bool)
    send_simulation_list = QtCore.pyqtSignal(list)
    send_material_list = QtCore.pyqtSignal(list)
    send_material_for_viewing = QtCore.pyqtSignal(object)
    send_view_array = QtCore.pyqtSignal(str, np.ndarray, str)  # simulation dict, array_slice, array_name, index, orientation
    send_view_array_bytescaled = QtCore.pyqtSignal(str, np.ndarray, str)  # simulation dict, array_slice, array_name, index, orientation
    send_view_array_slice = QtCore.pyqtSignal(str,str,  np.ndarray, int, int)  # simulation dict, array_slice, array_name, index, orientation
    send_view_sim_propeties = QtCore.pyqtSignal(dict)
    send_MC_ready_simulation = QtCore.pyqtSignal(dict, dict, list)


    send_proper_database_path = QtCore.pyqtSignal(QtCore.QUrl)

    def __init__(self, database_qurl, parent=None):
        super().__init__(parent)
        self.__db = None
        self.set_database(database_qurl)
        self.array_buffer = ArrayBuffer()
        self.array_buffer_size = 60

    @QtCore.pyqtSlot(QtCore.QUrl)
    def set_database(self, database_qurl):
        self.database_busy.emit(True)
        if self.__db:
            self.__db.close()
        fileinfo = QtCore.QFileInfo(database_qurl.toLocalFile())
        if fileinfo.isDir():
            folder = QtCore.QDir(fileinfo.absoluteFilePath())
            fname = 'database.h5'
            logger.info('Database path is a directory, setting database filename to: {0}'.format(fname))
        else:
            folder = fileinfo.dir()
            fname = fileinfo.fileName()

        msg = ''
        while not folder.exists():
            if not folder.mkpath(folder.absolutePath()):
                msg = 'Could not create folder: {0}, setting databasefolder to: {1}'.format(folder.absolutePath(), QtCore.QDir.current().absolutePath())
                logger.error(msg)
                folder = QtCore.QDir.current()

        path = QtCore.QFileInfo(folder.absolutePath() + folder.separator() + fname)

        logger.debug('Attemting to use database in {0}'.format(path.absoluteFilePath()))
        for_revert = self.__db
        try:
            self.__db = Database(path.absoluteFilePath())
        except OSError:    
            self.database_busy.emit(False)
            self.__db = for_revert            
            logger.debug('Failed to use database in {0}'.format(path.absoluteFilePath()))
        self.database_busy.emit(False)
        self.emit_material_list()
        self.emit_simulation_list()
        self.send_proper_database_path.emit(QtCore.QUrl.fromLocalFile(path.absoluteFilePath()))


    @QtCore.pyqtSlot()
    def emit_simulation_list(self):
        self.database_busy.emit(True)
        sims = self.__db.simulation_list()
        self.send_simulation_list.emit(sims)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot()
    def emit_material_list(self):
        self.database_busy.emit(True)
        mats = self.__db.material_list()
        self.send_material_list.emit(mats)
        self.database_busy.emit(False)


    @QtCore.pyqtSlot(str)
    def emit_material_for_viewing(self, mat_name):
        self.database_busy.emit(True)
        try:
            mat = self.__db.get_material(mat_name)
        except ValueError:
            pass
        else:
            self.send_material_for_viewing.emit(mat)
        self.database_busy.emit(False)


    @QtCore.pyqtSlot(dict, dict, bool)
    def add_simulation(self, properties, array_dict, overwrite):
        self.database_busy.emit(True)
        self.__db.add_simulation(properties, array_dict, overwrite)
        self.emit_simulation_list()
        self.database_busy.emit(False)

    @QtCore.pyqtSlot(str, str)
    def request_view_array(self, simulation_name, array_name):
        self.database_busy.emit(True)
        try:
            arr = self.__db.get_simulation_array(simulation_name, array_name)
        except ValueError:
            pass
        else:
            self.send_view_array.emit(simulation_name, arr, array_name)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot(str, str, float, float, bool)
    def request_view_array_bytescaled(self, simulation_name, array_name, amin, amax, vals_is_modifier):
        self.database_busy.emit(True)
        try:
            arr = self.__db.get_simulation_array_bytescaled(simulation_name, array_name, amin, amax, vals_is_modifier)
        except ValueError:
            pass
        else:
            self.send_view_array_bytescaled.emit(simulation_name, arr, array_name)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot(str, str, int, int)
    def request_view_array_slice(self, simulation_name, array_name, index, orientation):
        self.database_busy.emit(True)
        buffer_has_index, buffer_direction = self.array_buffer.is_slice_available(simulation_name, array_name, index, orientation)
        if buffer_has_index:
            logger.debug('Buffer has index {0} in {1}'.format(index, array_name))
            arr = self.array_buffer.get_slice(simulation_name, array_name, index, orientation)
            self.send_view_array_slice.emit(simulation_name, array_name, arr, index, orientation)

            if buffer_direction != 0:
                if buffer_direction > 0:
                    indices = np.arange(self.array_buffer_size, dtype=np.int) + index
                else:
                    indices = np.arange(self.array_buffer_size, dtype=np.int) + index - self.array_buffer_size + 1
                try:
                    arr = self.__db.get_simulation_array_slice(simulation_name, array_name, indices, orientation)
                except:
                    pass
                else:
                    logger.debug('Buffer need update, updating indices {0} in {1}'.format(indices, array_name))
                    self.array_buffer.set_buffer(arr, simulation_name, array_name, indices, orientation)

        else:
            try:
                arr = self.__db.get_simulation_array_slice(simulation_name, array_name, index, orientation)
            except:
                pass
            else:
                self.send_view_array_slice.emit(simulation_name, array_name, arr, index, orientation)

            indices = np.arange(self.array_buffer_size, dtype=np.int) + index - self.array_buffer_size // 2

            try:
                arr = self.__db.get_simulation_array_slice(simulation_name, array_name, indices, orientation)
            except:
                pass
            else:
                logger.debug('Buffer need update, updating indices {0} in {1}'.format(indices, array_name))
                self.array_buffer.set_buffer(arr, simulation_name, array_name, indices, orientation)

        self.database_busy.emit(False)

    @QtCore.pyqtSlot(list)
    def copy_simulation(self, names):
        for name in names:
            if isinstance(name, bytes):
                name = str(name, encoding='utf-8')
            self.database_busy.emit(True)
            self.__db.copy_simulation(name)
            self.database_busy.emit(False)
            self.emit_simulation_list()

    @QtCore.pyqtSlot(str)
    def request_simulation_properties(self, name):
        logger.debug('Request simulation metadata for {} from database.'.format(name))
        self.database_busy.emit(True)
        try:
            data = self.__db.get_simulation_metadata(name)
        except ValueError:
            logger.debug('Could not read metadata for simulation')
        else:
            self.send_view_sim_propeties.emit(data)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot(dict, bool, bool)
    def set_simulation_properties(self, propeties_dict, purge_volatiles=True, cancel_if_running=True):
        logger.debug('Request database to update simulation properties.')
        self.database_busy.emit(True)
        self.__db.set_simulation_metadata(propeties_dict, purge_volatiles, cancel_if_running)
        try:
            update_data = self.__db.get_simulation_metadata(propeties_dict.get('name', ''))
        except ValueError:
            logger.debug('Could not read metadata for simulation')
        else:
            self.send_view_sim_propeties.emit(update_data)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot(str, dict)
    def write_simulation_arrays(self, name, array_dict):
        logger.debug('Request database to write arrays.')
        self.database_busy.emit(True)
        for arr_name, array in array_dict.items():
            self.__db.set_simulation_array(name, array, arr_name)
        self.database_busy.emit(False)

    @QtCore.pyqtSlot()
    def request_MC_ready_simulation(self):
        self.database_busy.emit(True)
        try:
            props, arrays = self.__db.get_MCready_simulation()
        except ValueError:
            self.database_busy.emit(False)
            return
        else:
            pass

        try:
            materials = self.__db.get_materials(organic_only=False)
        except ValueError:
            self.database_busy.emit(False)
            logger.warning('Request to materials for a ready simulations failed for a mysterious reason.')
            return

        logger.debug('Emmitting signal request to run simulation {}'.format(props['name']))
        self.send_MC_ready_simulation.emit(props,arrays, materials)
        self.database_busy.emit(False)


    @QtCore.pyqtSlot(str)
    def remove_simulation(self, name):
        logger.debug('Attempting to remove simulation {}'.format(name))
        self.__db.remove_simulation(name)
        self.emit_simulation_list()



class Importer(QtCore.QObject):
    request_add_sim_to_database = QtCore.pyqtSignal(dict, dict, bool)
    running = QtCore.pyqtSignal(bool)

    def __init__(self, database_interface, parent=None):
        super().__init__(parent)
        self.__import_scaling = (1, 1, 1)
        self.request_add_sim_to_database.connect(database_interface.add_simulation)


    @QtCore.pyqtSlot(tuple)
    def set_import_scaling(self, im_scaling):
        self.__import_scaling = im_scaling


    @QtCore.pyqtSlot(list)
    def import_urls(self, qurl_list):
        self.running.emit(True)
        paths = [url.toLocalFile() for url in qurl_list]
        for props, arrays in import_ct_series(paths, import_scaling=self.__import_scaling):
            self.request_add_sim_to_database.emit(props, arrays, True)
        self.running.emit(False)

    @QtCore.pyqtSlot(list)
    def import_phantoms(self, qurl_list):
        self.running.emit(True)
        paths = [url.toLocalFile() for url in qurl_list]
        for props, arrays in read_phantoms(find_all_files(paths)):
            self.request_add_sim_to_database.emit(props, arrays, False)
        self.running.emit(False)

class Runner(QtCore.QThread):
    request_write_simulation_arrays = QtCore.pyqtSignal(str, dict)
    request_set_simulation_properties = QtCore.pyqtSignal(dict, bool, bool)
    request_runner_view_update = QtCore.pyqtSignal(np.ndarray, float, float, str, bool)

    start_timer = QtCore.pyqtSignal()
    def __init__(self, parent=None):
        super().__init__(parent)
        self.request_save = False
        self.timer = QtCore.QTimer()
        self.timer.setInterval(600000)

        self.started.connect(self.timer.start)
        self.timer.setSingleShot(False)
#        self.terminated.connect(self.timer.stop)
        self.finished.connect(self.timer.stop)

        self.timer.timeout.connect(self.set_request_save)
        self.mutex = QtCore.QMutex()
        self.simulation_properties = None
        self.simulation_arrays = None
        self.material_list = None

        self.kill_me = False
        self.is_running=False

    @QtCore.pyqtSlot()
    def set_request_save(self):
        self.mutex.lock()
        self.request_save = True
        self.mutex.unlock()
    @QtCore.pyqtSlot()
    def cancel_run(self):
        self.mutex.lock()
        self.kill_me = True
        self.mutex.unlock()

    def update_simulation_iteration(self, name, array_dict=None, exposure_number=None, progressbar_data=None, save=True):
        if self.kill_me:
            self.request_runner_view_update.emit(np.ones((3, 3)), 1, 1, '', False)
            self.mutex.lock()
            self.kill_me = False
            self.mutex.unlock()
            self.is_running=False
            self.terminated.emit()
            self.terminate()


#        if all([self.request_save, save, array_dict is not None,
#                exposure_number is not None]):
#            desc = {'name': name,
#                'start_at_exposure_no': exposure_number,
#                }
#            self.request_set_simulation_properties.emit(desc, False, False)
#            self.request_write_simulation_arrays.emit(name, array_dict)
#            self.mutex.lock()
#            self.request_save = False
#            self.mutex.unlock()
        if progressbar_data is not None:
            self.request_runner_view_update.emit(*progressbar_data)

    def run(self):
        self.is_running=True
        self.mutex.lock()
        self.kill_me = False





        simulation_properties = self.simulation_properties
        simulation_arrays = self.simulation_arrays
        material_list = self.material_list




        self.mutex.unlock()

        if simulation_properties is None or simulation_arrays is None:
            return
        if material_list is None:
            return

        self.request_set_simulation_properties.emit({'name': simulation_properties['name'],
                                                     'MC_running': True},
                                                     False, False)
        try:
            props_dict, arr_dict = ct_runner(material_list, simulation_properties,
                                             energy_imparted_to_dose_conversion=True,
                                             callback=self.update_simulation_iteration,
                                             **simulation_arrays)
        except MemoryError:
            logger.error('MEMORY ERROR: Could not run simulation {0}, memory to low. Try to increase dose matrix scaling or use 64 bit version of OpenDXMC'.format(simulation_properties['name']))
            simulation_properties['MC_finished'] = False
            simulation_properties['MC_running'] = False
            simulation_properties['MC_ready'] = False
            self.request_set_simulation_properties.emit(simulation_properties, True, False)
        except ValueError or AssertionError as e:
            raise e
            logger.error('UNKNOWN ERROR: Could not run simulation {0}'.format(simulation_properties['name']))
            simulation_properties['MC_finished'] = False
            simulation_properties['MC_running'] = False
            simulation_properties['MC_ready'] = False
            self.request_set_simulation_properties.emit(self.simulation_properties, True, False)
        else:
            self.request_write_simulation_arrays.emit(props_dict['name'], arr_dict)
            # generating doe array, watching memory
            energy_imparted = arr_dict.get('energy_imparted', None)
            density = arr_dict.get('density', None)
            c_factor = props_dict.get('conversion_factor_ctdiair')
            if c_factor == 0:
                c_factor = props_dict.get('conversion_factor_ctdiw')

            if energy_imparted is not None and density is not None:
                if c_factor > 0:
                    del arr_dict
                    try:
                        dose = (energy_imparted * c_factor) / (density * np.prod(props_dict['spacing'] * props_dict['scaling']))
                    except MemoryError:
                        logger.error('Memory error in generating dose array')
                    else:
                        self.request_write_simulation_arrays.emit(props_dict['name'], {'dose': dose})
            self.request_set_simulation_properties.emit(props_dict, False, False)
        self.simulation_properties = None
        self.simulation_arrays = None
        self.material_list = None
        self.request_save = False
        self.is_running=False


class RunManager(QtCore.QObject):
    mc_calculation_running = QtCore.pyqtSignal(bool)
    kill_runner = QtCore.pyqtSignal()
    def __init__(self, interface, progressbar, parent=None):
        super().__init__(parent)
        self.runner = Runner()
        self.kill_runner.connect(self.runner.cancel_run)
        interface.send_MC_ready_simulation.connect(self.run_simulation)
        self.runner.finished.connect(interface.request_MC_ready_simulation)
#        self.runner.terminated.connect(interface.request_MC_ready_simulation)
        self.runner.request_set_simulation_properties.connect(interface.set_simulation_properties)
        self.runner.request_write_simulation_arrays.connect(interface.write_simulation_arrays)
        self.runner.finished.connect(self.run_finished)
#        self.runner.terminated.connect(self.run_finished)
        self.runner.started.connect(self.run_started)
        self.runner.request_runner_view_update.connect(progressbar.set_data)
        self.current_simulation = ""


    @QtCore.pyqtSlot()
    def run_started(self):
        self.mc_calculation_running.emit(True)

    @QtCore.pyqtSlot(str)
    def cancel_run(self, name):
        if name == self.current_simulation:
            self.kill_runner.emit()
            self.current_simulation = ''

    @QtCore.pyqtSlot()
    def run_finished(self):
        self.current_simulation = ""
        self.mc_calculation_running.emit(False)

    @QtCore.pyqtSlot(dict, dict, list)
    def run_simulation(self, props, arrays, mat_list):
        logger.debug('Attemp to start MC thread')

        if not self.runner.is_running:
            self.runner.simulation_properties = props
            self.runner.simulation_arrays = arrays
            self.runner.material_list = mat_list
            self.current_simulation = props['name']
            self.runner.start()
#            self.runner.run()
            logger.debug('MC thread started')


class ListModel(QtCore.QAbstractListModel):

    request_data_list = QtCore.pyqtSignal()
    request_import_dicom = QtCore.pyqtSignal(list)
    request_import_phantom = QtCore.pyqtSignal(list)
    request_viewing = QtCore.pyqtSignal(str)
    request_copy_elements = QtCore.pyqtSignal(list)
    request_removal = QtCore.pyqtSignal(str)
    def __init__(self, interface, importer=None, phantom_importer=None, parent=None, simulations=False,
                 materials=False):
        super().__init__(parent)
        self.__data = []

        # connecting interface
        # outbound signals
        if simulations:
            self.request_data_list.connect(interface.emit_simulation_list)
            self.request_copy_elements.connect(interface.copy_simulation)
            self.request_removal.connect(interface.remove_simulation)
            if importer:
                self.request_import_dicom.connect(importer.import_urls)
            if phantom_importer:
                self.request_import_phantom.connect(phantom_importer.import_phantoms)
        elif materials:
            self.request_data_list.connect(interface.emit_material_list)


        # inbound signals
        if simulations:
            interface.send_simulation_list.connect(self.recive_data_list)
        elif materials:
            interface.send_material_list.connect(self.recive_data_list)

        # setting up
        self.request_data_list.emit()


    @QtCore.pyqtSlot(list)
    def recive_data_list(self, sims):
        self.layoutAboutToBeChanged.emit()
        self.__data = sims
        self.layoutChanged.emit()

    @QtCore.pyqtSlot(str)
    def element_activated(self, name):
        self.request_viewing.emit(name)

    @QtCore.pyqtSlot(str)
    def request_removal_emit(self, name):
        self.layoutAboutToBeChanged.emit()
        self.request_removal.emit(name)


    def rowCount(self, index):
        if not index.isValid():
            return len(self.__data)
        return 0

    def data(self, index, role):
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.__data):
            return None
        if role == QtCore.Qt.DisplayRole:
            return self.__data[row]
        elif role == QtCore.Qt.DecorationRole:
            pass
        elif role == QtCore.Qt.ToolTipRole:
            pass
        elif role == QtCore.Qt.SizeHintRole:
            pass
        elif role == QtCore.Qt.BackgroundRole:
            pass
        elif role == QtCore.Qt.ForegroundRole:
            pass
        return None

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        return str(section)

    def flags(self, index):
        if index.isValid():
            return QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsDropEnabled | QtCore.Qt.ItemIsDragEnabled
        return QtCore.Qt.ItemIsDropEnabled

    def mimeData(self, index_list):
        mimedata = QtCore.QMimeData()
        names = [self.__data[index.row()] for index in index_list if index.isValid()]
        if len(names) > 0:
            mimedata.setData('text/plain', (','.join(names)).encode('ascii'))
        return mimedata

    def mimeTypes(self):
#        return ['text/uri-list']
        return ['text/plain', 'text/uri-list']

    def dropMimeData(self, mimedata, action, row, column, index):
        if mimedata.hasUrls():
            urls = [u for u in mimedata.urls() if u.isLocalFile()]
            self.request_import_dicom.emit(urls)
            self.request_import_phantom.emit(urls)
            logger.debug(' '.join([u.toLocalFile() for u in urls]))
            return True
        elif mimedata.hasText():
            names = mimedata.data('text/plain').split(',')
            self.request_copy_elements.emit([str(n, encoding='ascii') for n in names])
        return False

    def supportedDropActions(self):
        return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction

    def __len__(self):
        return len(self.__data)

class ListView(QtWidgets.QListView):
    name_activated = QtCore.pyqtSignal(str)
    request_removal = QtCore.pyqtSignal(str)
    def __init__(self, parent=None, simulation=True):
        super().__init__(parent)
        self.tooltip = ""
        if simulation:
            self.setAcceptDrops(True)
            self.viewport().setAcceptDrops(True)
            self.setDropIndicatorShown(True)
            self.setDragDropMode(self.DragDrop)
            self.setDefaultDropAction(QtCore.Qt.CopyAction)
            self.tooltip = 'Drag DiCOM images or digital\nphantoms here to import'
            self.setToolTip(self.tooltip)

        else:
            self.setAcceptDrops(False)
            self.viewport().setAcceptDrops(False)
            self.setDropIndicatorShown(True)
            self.setDragDropMode(self.NoDragDrop)
            self.setDefaultDropAction(QtCore.Qt.CopyAction)

        self.activated.connect(self.activation_name)

    def setModel(self, model):
        try:
            self.name_activated.disconnect()
        except TypeError:
            pass
        super().setModel(model)
        self.name_activated.connect(model.element_activated)
        self.request_removal.connect(model.request_removal_emit)


    @QtCore.pyqtSlot(QtCore.QModelIndex)
    def activation_name(self, index):
        if index.isValid():
            self.name_activated.emit(index.data())

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        if event.key() == QtCore.Qt.Key_Delete:
            indices = self.selectedIndexes()
            for ind in indices:
                if ind.isValid():
                    self.request_removal.emit(ind.data())

    def paintEvent(self, event):
        super().paintEvent(event)
        if len(self.model()) == 0:
            painter = QtGui.QPainter(self.viewport())
            f = painter.font()
            f.setItalic(True)
            painter.setFont(f)
            painter.drawText(self.viewport().rect(),QtCore.Qt.AlignCenter | QtCore.Qt.TextWordWrap, self.tooltip)


class PropertiesEditModelItem(QtGui.QStandardItem):
    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = None

        init_val, dtype, volatile, editable, description, order, grouping = PROPETIES_DICT_TEMPLATE[key]
        self.dtype = dtype
        self.setEditable(editable)

        if dtype.type is np.bool_:
            self.setCheckable(True)
        self.update_data(value)


    def update_data(self, value):
        self.value = value
        if len(self.dtype.shape) > 0:
            val_t = " ".join([str(val) for val in value.astype(self.dtype.base.type)])
            self.setData(val_t, QtCore.Qt.DisplayRole)
        else:
            if self.dtype.type is np.bool_:
                if value:
                    c_val = QtCore.Qt.Checked
                else:
                    c_val = QtCore.Qt.Unchecked
                self.setData(c_val, QtCore.Qt.CheckStateRole)
            else:
                self.setData(value, QtCore.Qt.DisplayRole)


class PropertiesEditModel(QtGui.QStandardItemModel):
    request_properties_from_database = QtCore.pyqtSignal(str)
    request_write_properties_to_database = QtCore.pyqtSignal(dict, bool, bool)
    has_unsaved_changes = QtCore.pyqtSignal(bool)
    current_simulation_is_running = QtCore.pyqtSignal(bool)
    request_run_simulation = QtCore.pyqtSignal()
    request_cancel_simulation = QtCore.pyqtSignal(str)

    def __init__(self, database_interface, simulation_list_model, parent=None):
        super().__init__(parent)
        self.current_simulation = ""
        database_interface.send_view_sim_propeties.connect(self.set_simulation_properties)
        self.request_properties_from_database.connect(database_interface.request_simulation_properties)
        self.request_write_properties_to_database.connect(database_interface.set_simulation_properties)

        self.request_run_simulation.connect(database_interface.request_MC_ready_simulation)

        simulation_list_model.request_viewing.connect(self.set_simulation)

        self.validator = Validator()
        self.unsaved_items = {}

        propeties_dict, array_dict = self.validator.get_data()
        row = 0
        parent_items = {}

        values = list(PROPETIES_DICT_TEMPLATE_GROUPING.items())
        values.sort(key=lambda x:x[0])

        for key, value in values:
                parent_item = QtGui.QStandardItem(PROPETIES_DICT_TEMPLATE_GROUPING[key])
                parent_item.font().setBold(True)
                parent_item.setEditable(False)
                self.setItem(row, 0, parent_item)
                self.setItem(row, 1, QtGui.QStandardItem(' '))
                row += 1
                parent_items[key] = parent_item


        for key, value in propeties_dict.items():
            parent_ind = PROPETIES_DICT_TEMPLATE[key][6]
            if parent_ind in parent_items.keys():
                parent_item = parent_items[parent_ind]
            else:
                parent_item = QtGui.QStandardItem(PROPETIES_DICT_TEMPLATE_GROUPING.get(parent_ind, 'Unknown'))
                parent_item.font().setBold(True)
                parent_item.setEditable(False)
                self.setItem(row, 0, parent_item)
                self.setItem(row, 1, QtGui.QStandardItem(' '))
                row += 1
                parent_items[parent_ind] = parent_item

            item = QtGui.QStandardItem(PROPETIES_DICT_TEMPLATE[key][4])
            item.setEditable(False)
            parent_item.appendRow([item, PropertiesEditModelItem(key, value)])

        for parent_item in parent_items.values():
            parent_item.sortChildren(0)
#            parent_item.setChild(0, 0, item)
#            parent_item.setChild(0, 1, PropertiesEditModelItem(key, value))
#            self.setItem(row, 0, item)
#            self.setItem(row, 1, PropertiesEditModelItem(key, value))
#            row += 1

    def headerData(self, ind, orient, role):
        if (ind < 2) and (orient == QtCore.Qt.Horizontal):
            return ['Description', 'Value'][ind]
        return super().headerData(ind, orient, role)


    def data_item_iterator(self):
        for row in range(self.rowCount()):
            parent_item = self.item(row, 0)
            for child_row in range(parent_item.rowCount()):
                yield parent_item.child(child_row, 1)

    @QtCore.pyqtSlot(str)
    def set_simulation(self, name):
        self.current_simulation = name
        self.request_properties_from_database.emit(self.current_simulation)

    @QtCore.pyqtSlot(dict)
    def set_simulation_properties(self, data_dict):
        if data_dict['name'] != self.current_simulation:
            return

        self.validator.set_data(props=data_dict, reset=True)

        self.unsaved_items = {}
        for item in self.data_item_iterator():
            item.update_data(self.validator._props[item.key])
#        for row in range(self.rowCount()):
#            item = self.item(row, 1)
#            item.update_data(self.validator._props[item.key])
        self.test_unsaved_changes()
        self.current_simulation_is_running.emit(self.validator.MC_running)

    def test_unsaved_changes(self):
        keys_for_deletion = []
        for key, value in self.unsaved_items.items():
            if isinstance(value, np.ndarray):
                if np.sum(np.nonzero(value - getattr(self.validator, key))) == 0:
                    keys_for_deletion.append(key)
            else:
                if (value - getattr(self.validator, key)) == 0:
                    keys_for_deletion.append(key)
        for key in keys_for_deletion:
            del self.unsaved_items[key]

        self.has_unsaved_changes.emit(len(self.unsaved_items) > 0)


    def setData(self, index, value, role):
        if not index.isValid():
            return False
        if index.column() == 0:
            return super().setData(index, value, role)

        if role in [QtCore.Qt.DisplayRole, QtCore.Qt.EditRole, QtCore.Qt.CheckStateRole]:
            item = self.itemFromIndex(index)
            if not self.validator._pt[item.key][3]:

                return False
            if role == QtCore.Qt.CheckStateRole:
                value = value == QtCore.Qt.Checked

            try:
                setattr(self.validator, item.key, value)
            except AssertionError:
                return False
            else:
                if item.key not in self.unsaved_items:
                    self.unsaved_items[item.key] = item.value

#                for item in [self.item(i, 1) for i in range(self.rowCount())]:
                for item in self.data_item_iterator():
                    item.update_data(getattr(self.validator, item.key))
#
                self.test_unsaved_changes()
            return True
        return super().setData(index, value, role)

    @QtCore.pyqtSlot()
    def run_simulation(self):
        self.validator.MC_ready = True
        self.request_write_properties_to_database.emit(self.validator._props, True, True)
        self.request_run_simulation.emit()

    @QtCore.pyqtSlot()
    def apply_changes(self):
        self.request_write_properties_to_database.emit(self.validator._props, True, True)

    @QtCore.pyqtSlot()
    def reset_changes(self):
        self.validator.set_data(self.unsaved_items, reset=False)
        self.set_simulation_properties(self.validator.get_data()[0])

    @QtCore.pyqtSlot()
    def cancel_MC_run(self):
        if self.current_simulation == "":
            return

        self.validator.MC_ready = False
        self.validator.MC_running = False
        self.request_write_properties_to_database.emit(self.validator.get_data()[0], False, False)
        self.request_cancel_simulation.emit(self.current_simulation)



class PropertiesEditWidget(QtWidgets.QWidget):
    def __init__(self, database_interface, simulation_list_model, run_manager, parent=None):
        super().__init__(parent)
        layout = QtWidgets.QVBoxLayout()
        table = QtWidgets.QTreeView()
#        table.horizontalHeader().setResizeMode(QtWidgets.QHeaderView.Stretch)
        table.header().setResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        table.setAnimated(True)
        model = PropertiesEditModel(database_interface, simulation_list_model)
        self.model = model
        model.request_cancel_simulation.connect(run_manager.cancel_run)
        table.setModel(model)

        table.showColumn(0)
        table.showColumn(1)
        layout.addWidget(table)
        sub_layout = QtWidgets.QHBoxLayout()
        layout.addLayout(sub_layout)
        apply_button = QtWidgets.QPushButton()
        reset_button = QtWidgets.QPushButton()
        run_button = QtWidgets.QPushButton()
        cancel_button = QtWidgets.QPushButton()
        apply_button.setText('Apply')
        reset_button.setText('Reset')
        run_button.setText('Run')
        cancel_button.setText('Cancel')
        sub_layout.addWidget(reset_button)
        sub_layout.addWidget(apply_button)
        sub_layout.addWidget(run_button)
        sub_layout.addWidget(cancel_button)
        layout.setContentsMargins(0, 0, 0, 0)
        sub_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)


        reset_button.clicked.connect(model.reset_changes)
        cancel_button.clicked.connect(model.cancel_MC_run)
        apply_button.clicked.connect(model.apply_changes)
        run_button.clicked.connect(model.run_simulation)
        model.has_unsaved_changes.connect(reset_button.setEnabled)
        model.has_unsaved_changes.connect(apply_button.setEnabled)
        model.current_simulation_is_running.connect(cancel_button.setEnabled)
        model.current_simulation_is_running.connect(run_button.setDisabled)


class OrganDoseModel(QtCore.QAbstractTableModel):
    request_array = QtCore.pyqtSignal(str, str)
    request_array_slice = QtCore.pyqtSignal(str, str, int, int)
    hide_view = QtCore.pyqtSignal(bool)
    def __init__(self, database_interface, simulation_list_model, parent=None):
        super().__init__(parent)
        self.current_simulation = ""

        self.dose_z_lenght = []

        self.organ_array = None
        self.organ_material_map = None
        self.organ_map = None

        self._data = {}
        self._data_keys = []

        database_interface.send_view_array.connect(self.set_requested_array)
        database_interface.send_view_array_slice.connect(self.reload_slice)
        database_interface.send_view_sim_propeties.connect(self.set_simulation_properties)
        self.request_array.connect(database_interface.request_view_array)
        self.request_array_slice.connect(database_interface.request_view_array_slice)
        simulation_list_model.request_viewing.connect(self.set_simulation)


    def headerData(self, section, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            if section == 0:
                return 'Organ'
            elif section == 1:
                return 'Material'
            elif section == 2:
                return 'Dose [mGy/100mAs]'
        return None

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            r = index.row()
            c = index.column()
            if r >= len(self._data) or c > 2:
               return None
            if c == 2:
                if self._data[self._data_keys[r]][c+1] == 0:
                    return None
                return str(round(self._data[self._data_keys[r]][c] / self._data[self._data_keys[r]][c+1], 2))
            return self._data[self._data_keys[r]][c]
        return None
    def sort(self, column, order):
        self.layoutAboutToBeChanged.emit()
        reverse = True if order == QtCore.Qt.DescendingOrder else False
        if column in [0, 1]:
            self._data_keys.sort(key=lambda x: self._data[x][column], reverse=reverse)
        elif column == 2:
            self._data_keys.sort(key=lambda x: self._data[x][column]/self._data[x][column+1], reverse=reverse)
        self.layoutChanged.emit()

    def rowCount(self, index):
        return len(self._data)
    def columnCount(self, index):
        return 3

    @QtCore.pyqtSlot(str)
    def set_simulation(self, name):
        self.layoutAboutToBeChanged.emit()
        self.organ_array = None
        self.organ_material_map = None
        self.organ_map = None
        self._data = {}
        self._data_keys = []
        self.dose_z_lenght = []
        self.current_simulation = name
        self.layoutChanged.emit()

    @QtCore.pyqtSlot(dict)
    def set_simulation_properties(self, props_dict):
        self.hide_view.emit(True)
        if props_dict.get('name', "") != self.current_simulation:
            return
        self.scale = props_dict.get('scaling', np.ones(3))
        self.organ_array = None
        self.organ_material_map = None
        self.organ_map = None
        self.request_array.emit(self.current_simulation, 'organ_map')
        self.request_array.emit(self.current_simulation, 'organ_material_map')
        self.request_array.emit(self.current_simulation, 'organ')

    @QtCore.pyqtSlot(str, np.ndarray, str)
    def set_requested_array(self, name, array, array_name):
        if name != self.current_simulation:
            return
        if array_name == 'organ':
            self.organ_array = affine_transform(array,
                                 self.scale,
                                 output_shape=np.floor(np.array(array.shape)/self.scale).astype(np.int),
                                 cval=0, output=np.uint8, prefilter=True,
                                 order=0).astype(np.uint8)
        elif array_name == 'organ_map':
            self.organ_map = {array['organ'][i]: str(array['organ_name'][i], encoding='utf-8') for i in range(len(array))}
        elif array_name == 'organ_material_map':
            self.organ_material_map = {array['organ'][i]: str(array['material_name'][i], encoding='utf-8') for i in range(len(array))}

        if self.organ_array is not None and self.organ_map is not None and self.organ_material_map is not None:
            self.layoutAboutToBeChanged.emit()
            self._data = {}
            self._data_keys = []
            for key, value in self.organ_map.items():
                self._data[key] = [value, self.organ_material_map.get(key, 'Unknown'), 0, 0]
                self._data_keys.append(key)
            self.dose_z_lenght = list(range(self.organ_array.shape[0]))
#            self.dataChanged.emit(self.index(0, 0), self.index(len(self._data), 2))
            self.request_array_slice.emit(self.current_simulation, 'dose', 0, 0)
            self.layoutChanged.emit()
            self.hide_view.emit(False)


    @QtCore.pyqtSlot(str, str, np.ndarray, int, int)
    def reload_slice(self, name, array_name, arr, index, orientation):
        if orientation != 0 or array_name != 'dose' or name != self.current_simulation:
            return
        if self.organ_array is None:
            return
        if index in self.dose_z_lenght:
            for organ in np.unique(self.organ_array[index, :, :]):
                if organ not in self._data:
                    continue
#                self.layoutAboutToBeChanged.emit()
                ind_x, ind_y = np.nonzero(self.organ_array[index, :, :] == organ)
                self._data[organ][2] += arr[ind_x, ind_y].sum()
                self._data[organ][3] += len(ind_x)
#                import pdb
#                pdb.set_trace()
                model_index = self.index(self._data_keys.index(organ), 2)
#                self.layoutChanged.emit()
                self.dataChanged.emit(model_index, model_index)

            self.dose_z_lenght.remove(index)
        if len(self.dose_z_lenght) > 0:
            self.request_array_slice.emit(self.current_simulation, 'dose', self.dose_z_lenght[0], 0)


class OrganDoseView(QtWidgets.QTableView):
    def __init__(self, model, parent=None):
        super().__init__(parent)
        self.setModel(model)
        self.model().hide_view.connect(self.setHidden)
        self.horizontalHeader().setResizeMode(QtWidgets.QHeaderView.Stretch)
#        self.horizontalHeader().setVisible(True)
        self.setSortingEnabled(True)
        self.setDragEnabled(True)
        self.setAlternatingRowColors(True)
        self.setHidden(True)


    def copy_to_clipboard(self):
        indices = self.selectedIndexes()
        n_columns = self.model().columnCount(0)
        indices.sort(key = lambda x: x.row()*n_columns + x.column())

        c_row=0
        c_column=0

        html = "<table><tr>"
        txt = ""
        for index in indices:
            if index.isValid():
                if index.row() > c_row:
                    html += "</tr>"+"<tr>"
                    txt+="\n"
                if index.column() > c_column:
                    txt += "; "
                txt += str(index.data())
                html += "<td>" + str(index.data()) +"</td>"
                c_row = index.row()
                c_column = index.column()
        html += "</tr></table>"
        if len(html) > 0:
            mime=QtCore.QMimeData()
            mime.setHtml(html)
            mime.setText(txt)
            QtWidgets.qApp.clipboard().setMimeData(mime)


    def keyPressEvent(self, event):
        if event.matches(QtGui.QKeySequence.Copy):
            self.copy_to_clipboard()
            event.accept()
            return
        super().keyPressEvent(event)

